import mimetypes
from collections.abc import Sequence
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    Literal,
    Optional,
    Union,
    cast,
)

import orjson
import sqlalchemy

from edgy.core.db.context_vars import (
    CURRENT_INSTANCE,
    CURRENT_MODEL_INSTANCE,
    CURRENT_PHASE,
    EXPLICIT_SPECIFIED_VALUES,
)
from edgy.core.db.fields.base import BaseCompositeField
from edgy.core.db.fields.core import BigIntegerField, BooleanField, JSONField
from edgy.core.db.fields.factories import FieldFactory
from edgy.core.db.fields.types import ColumnDefinitionModel
from edgy.core.files.base import FieldFile, File
from edgy.core.files.storage import storages
from edgy.exceptions import FieldDefinitionError

if TYPE_CHECKING:
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.types import BaseModelType
    from edgy.core.files.storage import Storage

IGNORED = ["cls", "__class__", "kwargs", "generate_name_fn"]


class ConcreteFileField(BaseCompositeField):
    column_name: str = ""
    multi_process_safe: bool = True
    field_file_class: type[FieldFile]
    _generate_name_fn: Optional[
        Callable[[Optional["BaseModelType"], Union[File, BinaryIO], str, bool], str]
    ] = None

    def modify_input(self, name: str, kwargs: dict[str, Any]) -> None:
        # we are empty
        if name not in kwargs:
            return
        extracted_names: list[str] = [name, f"{name}_storage"]
        if self.with_size:
            extracted_names.append(f"{name}_size")
        if self.with_approval:
            extracted_names.append(f"{name}_approved")
        if self.with_metadata:
            extracted_names.append(f"{name}_metadata")
        to_add = {}
        for _name in extracted_names:
            if _name in kwargs:
                to_add[_name] = kwargs.pop(_name)
        kwargs[name] = to_add

    def generate_name_fn(
        self,
        instance: Optional["BaseModelType"],
        name: str,
        file: Union[File, BinaryIO],
        direct_name: bool,
    ) -> str:
        if self._generate_name_fn is None:
            return name
        return self._generate_name_fn(instance, file, name, direct_name)

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        """
        Inverse of clean. Transforms column(s) to a field for edgy.Model.
        Validation happens later.

        Args:
            field_name: the field name (can be different from name)
            value: the field value
        Kwargs:
            phase: the phase (set, creation, ...)

        """
        phase = CURRENT_PHASE.get()
        instance = CURRENT_INSTANCE.get()
        if (
            phase in {"post_update", "post_insert"}
            and instance is not None
            and isinstance(instance.__dict__.get(self.name), FieldFile)
        ):
            # use old one, when instance is no queryset
            field_instance_or_value: Any = cast(FieldFile, instance.__dict__[self.name])
        else:
            field_instance_or_value = value
        # unpack when not from db
        if isinstance(field_instance_or_value, dict) and not isinstance(
            field_instance_or_value.get(field_name), str
        ):
            field_instance_or_value = field_instance_or_value[field_name]
        if isinstance(field_instance_or_value, FieldFile):
            file_instance = field_instance_or_value
            if isinstance(value, dict):
                # update after post_insert/post_update, so just update some limited values
                # which does not affect operation
                if f"{field_name}_size" in value:
                    file_instance.size = value[f"{field_name}_size"]
                if value.get(f"{field_name}_metadata") is not None:
                    file_instance.metadata = value[f"{field_name}_metadata"]
                if value.get(f"{field_name}_approved") is not None:
                    file_instance.approved = value[f"{field_name}_approved"]
        else:
            # init, load, post_insert, post_update
            if isinstance(field_instance_or_value, dict):
                if phase == "set":
                    raise ValueError("Cannot set dict to FileField")
                file_instance = self.field_file_class(
                    self,
                    name=field_instance_or_value[field_name],
                    storage=field_instance_or_value.get(f"{field_name}_storage", self.storage),
                    size=field_instance_or_value.get(f"{field_name}_size"),
                    metadata=field_instance_or_value.get(f"{field_name}_metadata", {}),
                    approved=field_instance_or_value.get(
                        f"{field_name}_approved", not self.with_approval
                    ),
                    multi_process_safe=self.multi_process_safe,
                    change_removes_approval=self.with_approval,
                    generate_name_fn=partial(self.generate_name_fn, instance),
                )
            else:
                if instance is not None and self.name in instance.__dict__:
                    # use old one
                    file_instance = cast(FieldFile, instance.__dict__[self.name])
                else:
                    # not initialized yet
                    file_instance = self.field_file_class(
                        self,
                        multi_process_safe=self.multi_process_safe,
                        generate_name_fn=partial(self.generate_name_fn, instance),
                        storage=self.storage,
                        approved=not self.with_approval,
                        change_removes_approval=self.with_approval,
                    )
                # file creation if value is not None otherwise deletion
                file_instance.save(field_instance_or_value, delete_old=phase == "post_update")
        retdict: Any = {field_name: file_instance}
        if self.with_size:
            retdict[f"{field_name}_size"] = file_instance.size
        if self.with_approval:
            retdict[f"{field_name}_approved"] = file_instance.approved
        if self.with_metadata:
            metadata_result: Any = file_instance.metadata
            field = self.owner.meta.fields[f"{field_name}_metadata"]
            if field.field_type is str:
                metadata_result = orjson.dumps(metadata_result).decode("utf8")
            retdict[f"{field_name}_metadata"] = metadata_result
        return retdict  # type: ignore

    def get_columns(self, field_name: str) -> Sequence[sqlalchemy.Column]:
        model = ColumnDefinitionModel.model_validate(self, from_attributes=True)
        column_name = self.column_name or field_name
        return [
            sqlalchemy.Column(
                key=field_name,
                type_=model.column_type,
                name=column_name,
                **model.model_dump(by_alias=True, exclude_none=True, exclude={"column_name"}),
            ),
            sqlalchemy.Column(
                key=f"{field_name}_storage",
                name=f"{column_name}_storage",
                type_=sqlalchemy.String(length=20, collation=self.column_type.collation),
                default=self.storage.name,
            ),
        ]

    def get_embedded_fields(
        self, name: str, fields: dict[str, "BaseFieldType"]
    ) -> dict[str, "BaseFieldType"]:
        retdict: dict[str, Any] = {}
        column_name = self.column_name or name
        # TODO: use embed_field
        if self.with_size:
            size_name = f"{name}_size"
            if size_name not in fields:
                retdict[size_name] = BigIntegerField(
                    ge=0,
                    null=self.null,
                    exclude=True,
                    read_only=True,
                    name=size_name,
                    column_name=f"{column_name}_size",
                    owner=self.owner,
                )
        if self.with_approval:
            approval_name = f"{name}_approved"
            if approval_name not in fields:
                retdict[approval_name] = BooleanField(
                    null=False,
                    default=False,
                    exclude=True,
                    column_name=f"{column_name}_ok",
                    name=approval_name,
                    owner=self.owner,
                )
        if self.with_metadata:
            metadata_name = f"{name}_metadata"
            if metadata_name not in fields:
                retdict[metadata_name] = JSONField(
                    null=False,
                    column_name=f"{column_name}_mname",
                    name=metadata_name,
                    owner=self.owner,
                    default=dict,
                )
        return retdict

    def get_composite_fields(self) -> dict[str, "BaseFieldType"]:
        field_names: list[str] = [self.name]
        if self.with_size:
            field_names.append(f"{self.name}_size")
        if self.with_approval:
            field_names.append(f"{self.name}_approved")
        if self.with_metadata:
            field_names.append(f"{self.name}_metadata")
        return {name: self.owner.meta.fields[name] for name in field_names}

    async def post_save_callback(
        self, value: FieldFile, instance: "BaseModelType", force_insert: bool
    ) -> None:
        await value.execute_operation(nodelete_old=force_insert)
        # cleanup temp file
        value.close(keep_size=True)

    async def post_delete_callback(self, value: FieldFile, instance: "BaseModelType") -> None:
        value.delete(instant=True)


class FileField(FieldFactory):
    field_type = Any
    field_bases = (ConcreteFileField,)

    def __new__(  # type: ignore
        cls,
        storage: Union[str, "Storage", None] = None,
        with_size: bool = True,
        with_metadata: bool = True,
        with_approval: bool = False,
        extract_mime: Union[bool, Literal["approved_only"]] = True,
        mime_use_magic: bool = False,
        field_file_class: type[FieldFile] = FieldFile,
        generate_name_fn: Optional[
            Callable[[Optional["BaseModelType"], Union[File, BinaryIO], str, bool], str]
        ] = None,
        **kwargs: Any,
    ) -> "BaseFieldType":
        if not storage:
            storage = storages["default"]
        elif isinstance(storage, str):
            storage = storages[storage]

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in IGNORED},
        }
        return super().__new__(cls, _generate_name_fn=generate_name_fn, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        if kwargs.get("mime_use_magic"):
            try:
                import magic  # noqa: F401
            except ImportError:
                raise FieldDefinitionError(
                    "python-magic library is missing. Cannot use mime_use_magic parameter"
                ) from None

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.String(
            length=kwargs.get("max_length", 255), collation=kwargs.get("collation")
        )

    @classmethod
    def extract_metadata(
        cls, field_obj: "BaseFieldType", field_name: str, field_file: FieldFile
    ) -> dict[str, Any]:
        data: dict[str, Any] = {}
        if field_obj.extract_mime and (
            field_file.approved or field_obj.extract_mime != "approved_only"
        ):
            if getattr(field_obj, "mime_use_magic", False):
                from magic import Magic

                magic = Magic(mime=True)
                # only open, not close, done later
                data["mime"] = magic.from_buffer(field_file.open("rb").read(2048))
            else:
                data["mime"] = mimetypes.guess_type(field_file.name)[0]
        return data

    @classmethod
    def clean(
        cls,
        field_obj: "BaseFieldType",
        field_name: str,
        value: Union[FieldFile, str, dict],
        for_query: bool = False,
        original_fn: Any = None,
    ) -> dict[str, Any]:
        """
        Validates a value and transform it into columns which can be used for querying and saving.

        Args:
            field_name: the field name (can be different from name)
            value: the field value
        Kwargs:
            for_query: is used for querying. Should have all columns used for querying set.
        """
        assert field_obj.owner
        # unpack
        if isinstance(value, dict) and field_name in value:
            if for_query:
                if isinstance(value[field_name], (FieldFile, str, type(None))):
                    value = value[field_name]
            else:
                # to_model is assumed to be called already
                phase = CURRENT_PHASE.get()
                explicit_values = EXPLICIT_SPECIFIED_VALUES.get()
                if isinstance(value[field_name], FieldFile):
                    value = value[field_name]
                    if (
                        phase == "prepare_insert"
                        and explicit_values is not None
                        and field_name not in explicit_values
                    ):
                        cast(FileField, value).save(
                            cast(FileField, value).to_file(), delete_old=False
                        )
                else:
                    instance = CURRENT_MODEL_INSTANCE.get()
                    assert instance is not None, "No model instance found"
                    # use model instance
                    to_save = value[field_name]
                    value = cast("FieldFile", getattr(instance, field_name))
                    # set the file and file name
                    value.save(to_save, delete_old=phase == "prepare_update")

        # handle None
        if value is None:
            nulldict: dict[str, Any] = {
                field_name: None,
            }
            nulldict[f"{field_name}_storage"] = None
            if not for_query:
                if field_obj.with_approval:
                    nulldict[f"{field_name}_approved"] = False
                if field_obj.with_size:
                    nulldict[f"{field_name}_size"] = None
                if field_obj.with_metadata:
                    nulldict[f"{field_name}_metadata"] = {}
            return nulldict

        if for_query:
            if isinstance(value, str):
                return {field_name: value}
            assert isinstance(value, FieldFile)
            query_dict: dict[str, Any] = {
                field_name: value.name,
            }
            query_dict[f"{field_name}_storage"] = value.storage.name
            return query_dict
        else:
            if not isinstance(value, FieldFile):
                raise ValueError(f"invalid value for for_query=False: {value} ({value!r})")
            retdict: dict[str, Any] = {
                field_name: value.name,
            }
            retdict[f"{field_name}_storage"] = value.storage.name
            if field_obj.with_approval:
                retdict[f"{field_name}_approved"] = value.approved
            if field_obj.with_size:
                retdict[f"{field_name}_size"] = value.size if value else None
            if field_obj.with_metadata:
                metadata_result: Any = (
                    cls.extract_metadata(field_obj, field_name=field_name, field_file=value)
                    if value
                    else {}
                )
                field = field_obj.owner.meta.fields[f"{field_name}_metadata"]
                # in case the field was swapped out for a text field
                if field.field_type is str:
                    metadata_result = orjson.dumps(metadata_result).decode("utf8")
                retdict[f"{field_name}_metadata"] = metadata_result
            return retdict



import warnings
from base64 import b32encode
from functools import lru_cache
from hashlib import blake2b
from typing import TYPE_CHECKING

from edgy.exceptions import DatabaseNotConnectedWarning

if TYPE_CHECKING:
    from edgy.core.connection.database import Database


def check_db_connection(db: "Database", stacklevel: int = 3) -> None:
    if not db.is_connected:
        # with force_rollback the effects are even worse, so fail
        if db.force_rollback:
            raise RuntimeError("db is not connected.")
        # db engine will be created and destroyed afterwards
        warnings.warn(
            "Database not connected. Executing operation is inperformant.",
            DatabaseNotConnectedWarning,
            stacklevel=stacklevel,
        )


@lru_cache(512, typed=False)
def _hash_tablekey(tablekey: str, prefix: str) -> str:
    tablehash = (
        b32encode(blake2b(f"{tablekey}_{prefix}".encode(), digest_size=16).digest())
        .decode()
        .rstrip("=")
    )

    return f"_join_{tablehash}"


def hash_tablekey(*, tablekey: str, prefix: str) -> str:
    """
    For temporary aliases like joins.
    Columns can be extracted from joins by adding to the hash `_<column name>`.
    """
    if not prefix:
        return tablekey
    return _hash_tablekey(tablekey, prefix)



from __future__ import annotations

import contextlib
import copy
import inspect
import warnings
from abc import ABCMeta
from collections import UserDict, deque
from collections.abc import Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Optional,
    Union,
    cast,
    get_origin,
)

import sqlalchemy
from pydantic._internal._model_construction import ModelMetaclass

from edgy.core import signals as signals_module
from edgy.core.connection.registry import Registry
from edgy.core.db import fields as edgy_fields
from edgy.core.db.datastructures import Index, UniqueConstraint
from edgy.core.db.fields.base import PKField, RelationshipField
from edgy.core.db.fields.foreign_keys import BaseForeignKeyField
from edgy.core.db.fields.types import BaseFieldType
from edgy.core.db.models.managers import BaseManager
from edgy.core.db.models.utils import build_pkcolumns, build_pknames
from edgy.core.utils.functional import extract_field_annotations_and_defaults
from edgy.exceptions import ImproperlyConfigured, TableBuildError

if TYPE_CHECKING:
    from edgy.core.db.models import Model
    from edgy.core.db.models.types import BaseModelType

_empty_dict: dict[str, Any] = {}
_empty_set: frozenset[Any] = frozenset()


class Fields(UserDict, dict[str, BaseFieldType]):
    """Smart wrapper which tries to prevent invalidation as far as possible"""

    def __init__(self, meta: MetaInfo, data: Optional[dict[str, BaseFieldType]] = None):
        self.meta = meta
        super().__init__(data)

    def add_field_to_meta(self, name: str, field: BaseFieldType) -> None:
        if not self.meta._fields_are_initialized:
            return
        if hasattr(field, "__get__"):
            self.meta.special_getter_fields.add(name)
        if getattr(field, "exclude", False):
            self.meta.excluded_fields.add(name)
        if getattr(field, "secret", False):
            self.meta.secret_fields.add(name)
        if hasattr(field, "modify_input"):
            self.meta.input_modifying_fields.add(name)
        if hasattr(field, "post_save_callback"):
            self.meta.post_save_fields.add(name)
        if hasattr(field, "pre_save_callback"):
            self.meta.pre_save_fields.add(name)
        if hasattr(field, "post_delete_callback"):
            self.meta.post_delete_fields.add(name)
        if isinstance(field, BaseForeignKeyField):
            self.meta.foreign_key_fields.add(name)
        if isinstance(field, RelationshipField):
            self.meta.relationship_fields.add(name)

    def discard_field_from_meta(self, name: str) -> None:
        if self.meta._fields_are_initialized:
            for field_attr in _field_sets_to_clear:
                getattr(self.meta, field_attr).discard(name)

    def __getitem__(self, name: str) -> BaseFieldType:
        return cast(BaseFieldType, self.data[name])

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __setitem__(self, name: str, value: BaseFieldType) -> None:
        if name in self.data:
            self.discard_field_from_meta(name)
        self.data[name] = value
        self.add_field_to_meta(name, value)
        if self.meta.model is not None:
            self.meta.model.model_fields[name] = value  # type: ignore
        self.meta.invalidate(clear_class_attrs=True)

    def __delitem__(self, name: str) -> None:
        if self.data.pop(name, None) is not None:
            self.discard_field_from_meta(name)
            self.meta.invalidate(clear_class_attrs=True)


class FieldToColumns(UserDict, dict[str, Sequence["sqlalchemy.Column"]]):
    def __init__(self, meta: MetaInfo):
        self.meta = meta
        super().__init__()

    def __getitem__(self, name: str) -> Sequence[sqlalchemy.Column]:
        if name in self.data:
            return cast(Sequence["sqlalchemy.Column"], self.data[name])
        field = self.meta.fields[name]
        result = self.data[name] = field.get_columns(name)
        return result

    def __setitem__(self, name: str, value: Any) -> None:
        raise Exception("Cannot set item here")

    def __iter__(self) -> Any:
        self.meta.columns_to_field.init()
        return super().__iter__()

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key: str) -> bool:
        try:
            self[key]
            return True
        except KeyError:
            return False


class FieldToColumnNames(FieldToColumns, dict[str, frozenset[str]]):
    def __getitem__(self, name: str) -> frozenset[str]:
        if name in self.data:
            return cast(frozenset[str], self.data[name])
        column_names = frozenset(column.key for column in self.meta.field_to_columns[name])
        result = self.data[name] = column_names
        return result


class ColumnsToField(UserDict, dict[str, str]):
    def __init__(self, meta: MetaInfo):
        self.meta = meta
        self._init = False
        super().__init__()

    def init(self) -> None:
        if not self._init:
            self._init = True
            _columns_to_field: dict[str, str] = {}
            for field_name in self.meta.fields:
                # init structure
                column_names = self.meta.field_to_column_names[field_name]
                for column_name in column_names:
                    if column_name in _columns_to_field:
                        raise ValueError(
                            f"column collision: {column_name} between field {field_name} and {_columns_to_field[column_name]}"
                        )
                    _columns_to_field[column_name] = field_name
            self.data.update(_columns_to_field)

    def __getitem__(self, name: str) -> str:
        self.init()
        return cast(str, super().__getitem__(name))

    def __setitem__(self, name: str, value: Any) -> None:
        raise Exception("Cannot set item here")

    def __contains__(self, name: str) -> bool:
        self.init()
        return super().__contains__(name)

    def __iter__(self) -> Any:
        self.init()
        return super().__iter__()


_trigger_attributes_MetaInfo = {
    "field_to_columns",
    "field_to_column_names",
    "foreign_key_fields",
    "columns_to_field",
    "special_getter_fields",
    "input_modifying_fields",
    "post_save_fields",
    "post_delete_fields",
    "excluded_fields",
    "secret_fields",
    # not used here in code but usefully for third-party code
    "relationship_fields",
}

_field_sets_to_clear: set[str] = {
    attr for attr in _trigger_attributes_MetaInfo if attr.endswith("_fields")
}


class MetaInfo:
    __slots__ = (
        "abstract",
        "inherit",
        "fields",
        "registry",
        "tablename",
        "unique_together",
        "indexes",
        "model",
        "managers",
        "multi_related",
        "signals",
        "input_modifying_fields",
        "pre_save_fields",
        "post_save_fields",
        "post_delete_fields",
        "foreign_key_fields",
        "field_to_columns",
        "field_to_column_names",
        "columns_to_field",
        "special_getter_fields",
        "excluded_fields",
        "secret_fields",
        "relationship_fields",
        "_fields_are_initialized",
    )
    _include_dump = (
        *filter(
            lambda x: x
            not in {
                "field_to_columns",
                "field_to_column_names",
                "columns_to_field",
                "_fields_are_initialized",
            },
            __slots__,
        ),
        "pk",
    )

    fields: Fields
    field_to_columns: FieldToColumns
    field_to_column_names: FieldToColumnNames
    columns_to_field: ColumnsToField

    def __init__(self, meta: Any = None, **kwargs: Any) -> None:
        self._fields_are_initialized = False
        self.model: Optional[type[BaseModelType]] = None
        #  Difference between meta extraction and kwargs: meta attributes are copied
        self.abstract: bool = getattr(meta, "abstract", False)
        # for embedding
        self.inherit: bool = getattr(meta, "inherit", True)
        self.registry: Union[Registry, Literal[False], None] = getattr(meta, "registry", None)
        self.tablename: Optional[str] = getattr(meta, "tablename", None)
        self.unique_together: Any = getattr(meta, "unique_together", None)
        self.indexes: Any = getattr(meta, "indexes", None)
        self.signals = signals_module.Broadcaster(getattr(meta, "signals", None) or {})
        self.signals.set_lifecycle_signals_from(signals_module, overwrite=False)
        self.fields = {**getattr(meta, "fields", _empty_dict)}  # type: ignore
        self.managers: dict[str, BaseManager] = {**getattr(meta, "managers", _empty_dict)}
        self.multi_related: set[tuple[str, str]] = {*getattr(meta, "multi_related", _empty_set)}
        self.load_dict(kwargs)

    @property
    def pk(self) -> Optional[PKField]:
        return cast(Optional[PKField], self.fields.get("pk"))

    @property
    def fields_mapping(self) -> dict[str, BaseFieldType]:
        warnings.warn(
            "'fields_mapping' has been deprecated, use 'fields' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.fields

    @property
    def is_multi(self) -> bool:
        warnings.warn(
            "`is_multi` is deprecated. Use bool(meta.multi_related) instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return bool(self.multi_related)

    @property
    def parents(self) -> list[Any]:
        warnings.warn(
            "`parents` is deprecated and will be removed without replacement.",
            DeprecationWarning,
            stacklevel=2,
        )
        return [parent for parent in self.model.__bases__ if isinstance(parent, BaseModelMeta)]

    def model_dump(self) -> dict[Any, Any]:
        return {k: getattr(self, k, None) for k in self._include_dump}

    def load_dict(self, values: dict[str, Any]) -> None:
        """
        Loads the metadata from a dictionary.
        You may want to overload it to create hooks to ensure types.
        """
        for key, value in values.items():
            # we want triggering invalidate in case it is fields
            setattr(self, key, value)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "fields":
            value = Fields(self, value)
        super().__setattr__(name, value)
        if name == "fields":
            self.invalidate()

    def __getattribute__(self, name: str) -> Any:
        # lazy execute
        if name in _trigger_attributes_MetaInfo and not self._fields_are_initialized:
            self.init_fields_mapping()
        return super().__getattribute__(name)

    def init_fields_mapping(self) -> None:
        self.special_getter_fields: set[str] = set()
        self.excluded_fields: set[str] = set()
        self.secret_fields: set[str] = set()
        self.input_modifying_fields: set[str] = set()
        self.post_save_fields: set[str] = set()
        self.pre_save_fields: set[str] = set()
        self.post_delete_fields: set[str] = set()
        self.foreign_key_fields: set[str] = set()
        self.relationship_fields: set[str] = set()
        self.field_to_columns = FieldToColumns(self)
        self.field_to_column_names = FieldToColumnNames(self)
        self.columns_to_field = ColumnsToField(self)
        self._fields_are_initialized = True
        for key, field in self.fields.items():
            self.fields.add_field_to_meta(key, field)

    def invalidate(self, clear_class_attrs: bool = True, invalidate_fields: bool = True) -> None:
        if invalidate_fields and self._fields_are_initialized:
            # prevent cycles and mem-leaks
            for attr in (
                "field_to_columns",
                "field_to_column_names",
                "columns_to_field",
            ):
                with contextlib.suppress(AttributeError):
                    delattr(self, attr)
            self._fields_are_initialized = False
        if self.model is None:
            return
        if clear_class_attrs:
            for attr in ("_table", "_pknames", "_pkcolumns", "_db_schemas", "__proxy_model__"):
                with contextlib.suppress(AttributeError):
                    delattr(self.model, attr)

    def full_init(self, init_column_mappers: bool = True, init_class_attrs: bool = True) -> None:
        if not self._fields_are_initialized:
            self.init_fields_mapping()
        if init_column_mappers:
            self.columns_to_field.init()
        if init_class_attrs:
            for attr in ("table", "pknames", "pkcolumns", "proxy_model"):
                getattr(self.model, attr)

    def get_columns_for_name(self, name: str) -> Sequence[sqlalchemy.Column]:
        if name in self.field_to_columns:
            return self.field_to_columns[name]
        elif self.model and name in self.model.table.columns:
            return (self.model.table.columns[name],)
        else:
            return cast(Sequence["sqlalchemy.Column"], _empty_set)


def get_model_registry(
    bases: tuple[type, ...], meta_class: Optional[Union[object, MetaInfo]] = None
) -> Union[Registry, None, Literal[False]]:
    """
    When a registry is missing from the Meta class, it should look up for the bases
    and obtain the first found registry.
    """
    if meta_class is not None:
        direct_registry: Union[Registry, None, Literal[False]] = getattr(
            meta_class, "registry", None
        )
        if direct_registry is not None:
            return direct_registry

    for base in bases:
        meta: MetaInfo = getattr(base, "meta", None)
        # now check meta
        if meta is None:
            continue
        found_registry: Union[Registry, None, Literal[False]] = getattr(meta, "registry", None)

        if found_registry is not None:
            return found_registry
    return None


def _handle_annotations(base: type, base_annotations: dict[str, Any]) -> None:
    for parent in base.__mro__[1:]:
        _handle_annotations(parent, base_annotations)
    if hasattr(base, "__init_annotations__") and base.__init_annotations__:
        base_annotations.update(base.__init_annotations__)
    elif hasattr(base, "__annotations__") and base.__annotations__:
        # python 3.9 has no get_annotations
        base_annotations.update(base.__annotations__)


def handle_annotations(
    bases: tuple[type, ...], base_annotations: dict[str, Any], attrs: Any
) -> dict[str, Any]:
    """
    Handles and copies some of the annotations for
    initialisation.
    """
    for base in bases:
        _handle_annotations(base, base_annotations)

    annotations: dict[str, Any] = (
        copy.copy(attrs["__init_annotations__"])
        if "__init_annotations__" in attrs
        else copy.copy(attrs["__annotations__"])
    )
    annotations.update(base_annotations)
    return annotations


_occluded_sentinel = object()


def _extract_fields_and_managers(base: type, attrs: dict[str, Any]) -> None:
    from edgy.core.db.fields.composite_field import CompositeField

    meta: Union[MetaInfo, None] = getattr(base, "meta", None)
    if not meta:
        # Mixins and other classes
        # Note: from mixins BaseFields and BaseManagers are imported despite inherit=False until a model in the
        # hierarchy uses them
        # Here is _occluded_sentinel not overwritten
        for key, value in inspect.getmembers(base):
            if key not in attrs:
                if isinstance(value, BaseFieldType):
                    attrs[key] = value
                elif isinstance(value, BaseManager):
                    attrs[key] = value.__class__()
                elif isinstance(value, BaseModelMeta):
                    attrs[key] = CompositeField(
                        inner_fields=value,
                        prefix_embedded=f"{key}_",
                        inherit=value.meta.inherit,
                        name=key,
                        owner=value,
                    )
            elif attrs[key] is _occluded_sentinel:
                # when occluded only include if inherit is True
                if isinstance(value, BaseFieldType) and value.inherit:
                    attrs[key] = value
                elif isinstance(value, BaseManager) and value.inherit:
                    attrs[key] = value.__class__()
                elif isinstance(value, BaseModelMeta) and value.meta.inherit:
                    attrs[key] = CompositeField(
                        inner_fields=value,
                        prefix_embedded=f"{key}_",
                        inherit=value.meta.inherit,
                        name=key,
                        owner=value,
                    )

    else:
        # abstract classes
        for key, value in meta.fields.items():
            if key not in attrs:
                # when abstract or inherit passthrough
                if meta.abstract or value.inherit:
                    attrs[key] = value
                    assert value.owner is not None
                else:
                    attrs[key] = _occluded_sentinel
            elif attrs[key] is _occluded_sentinel and value.inherit:
                # when occluded only include if inherit is True
                attrs[key] = value
                assert value.owner is not None
        for key, value in meta.managers.items():
            if key not in attrs:
                # when abstract or inherit passthrough
                if meta.abstract or value.inherit:
                    attrs[key] = value
                else:
                    attrs[key] = _occluded_sentinel
            elif attrs[key] is _occluded_sentinel and value.inherit:
                # when occluded only include if inherit is True
                attrs[key] = value

    # from strongest to weakest
    for parent in base.__mro__[1:]:
        _extract_fields_and_managers(parent, attrs)


def extract_fields_and_managers(
    bases: Sequence[type], attrs: Optional[dict[str, Any]] = None
) -> dict[str, Any]:
    """
    Search for fields and managers and return them.

    Managers are copied.

    Note: managers and fields with inherit=False are still extracted from mixins as long there is no intermediate model
    """

    from edgy.core.db.fields.composite_field import CompositeField

    attrs = {} if attrs is None else {**attrs}
    # order is important
    for key in list(attrs.keys()):
        value = attrs[key]
        if isinstance(value, BaseModelMeta):
            value = attrs[key]
            attrs[key] = CompositeField(
                inner_fields=value,
                prefix_embedded=f"{key}_",
                inherit=value.meta.inherit,
                owner=value,
            )
    for base in bases:
        _extract_fields_and_managers(base, attrs)
    # now remove sentinels
    for key in list(attrs.keys()):
        value = attrs[key]
        if value is _occluded_sentinel:
            attrs.pop(key)
    return attrs


class BaseModelMeta(ModelMetaclass, ABCMeta):
    __slots__ = ()

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, Any],
        meta_info_class: type[MetaInfo] = MetaInfo,
        skip_registry: bool = False,
        **kwargs: Any,
    ) -> Any:
        fields: dict[str, BaseFieldType] = {}
        managers: dict[str, BaseManager] = {}
        meta_class: object = attrs.get("Meta", type("Meta", (), {}))
        base_annotations: dict[str, Any] = {}
        has_explicit_primary_key = False
        is_abstract: bool = getattr(meta_class, "abstract", False)
        has_parents = any(isinstance(parent, BaseModelMeta) for parent in bases)

        # Extract the custom Edgy Fields in a pydantic format.
        attrs, model_fields = extract_field_annotations_and_defaults(attrs)
        # ensure they are clean
        attrs.pop("_pkcolumns", None)
        attrs.pop("_pknames", None)
        attrs.pop("_table", None)

        # Extract fields and managers and include them in attrs
        attrs = extract_fields_and_managers(bases, attrs)

        for key, value in attrs.items():
            if isinstance(value, BaseFieldType):
                if key == "pk" and not isinstance(value, PKField):
                    raise ImproperlyConfigured(
                        f"Cannot add a field named pk to model {name}. Protected name."
                    )
                # make sure we have a fresh copy where we can set the owner and overwrite methods
                value = copy.copy(value)
                if value.factory is not None:
                    value.factory.repack(value)
                if value.primary_key:
                    has_explicit_primary_key = True
                # set as soon as possible the field_name
                value.name = key

                # add fields and non BaseRefForeignKeyField to fields
                # The BaseRefForeignKeyField is actually not a normal SQL Foreignkey
                # It is an Edgy specific operation that creates a reference to a ForeignKey
                # That is why is not stored as a normal FK but as a reference but
                # stored also as a field to be able later or to access anywhere in the model
                # and use the value for the creation of the records via RefForeignKey.
                # saving a reference foreign key.
                # We split the keys (store them) in different places to be able to easily maintain and
                #  what is what.
                fields[key] = value
                model_fields[key] = value
            elif isinstance(value, BaseManager):
                value = copy.copy(value)
                value.name = key
                managers[key] = value

        if not is_abstract:
            # the order is important because it reflects the inheritance order
            # from the strongest to the weakest
            fieldnames_to_check = deque(fields.keys())
            while fieldnames_to_check:
                # pop the weakest next field
                field_name = fieldnames_to_check.pop()
                field = fields[field_name]
                embedded_fields = field.get_embedded_fields(field_name, fields)
                if embedded_fields:
                    for sub_field_name, sub_field in embedded_fields.items():
                        if sub_field_name == "pk":
                            raise ValueError("sub field uses reserved name pk")

                        if sub_field_name in fields and fields[sub_field_name].owner is None:
                            raise ValueError(f"sub field name collision: {sub_field_name}")
                        # set as soon as possible the field_name
                        sub_field.name = sub_field_name
                        if sub_field.primary_key:
                            has_explicit_primary_key = True
                        # insert as stronger element
                        fieldnames_to_check.appendleft(sub_field_name)
                        fields[sub_field_name] = sub_field
                        if not sub_field.exclude:
                            model_fields[sub_field_name] = sub_field
            # Handle with multiple primary keys and auto generated field if no primary key is provided
            if not is_abstract and has_parents and not has_explicit_primary_key:
                if "id" not in fields:
                    if attrs.get("__reflected__", False):
                        raise ImproperlyConfigured(
                            f"Cannot create model {name}. No primary key found and reflected."
                        )
                    else:
                        model_fields["id"] = fields["id"] = edgy_fields.BigIntegerField(  # type: ignore
                            primary_key=True,
                            autoincrement=True,
                            inherit=False,
                            name="id",
                        )
                if not isinstance(fields["id"], BaseFieldType) or not fields["id"].primary_key:
                    raise ImproperlyConfigured(
                        f"Cannot create model {name} without explicit primary key if field 'id' is already present."
                    )

        for field_name in fields:
            attrs.pop(field_name, None)

        for manager_name in managers:
            attrs.pop(manager_name, None)
        attrs["meta"] = meta = meta_info_class(
            meta_class,
            fields=fields,
            managers=managers,
        )
        del fields
        if is_abstract or not meta.fields:
            meta.abstract = True

        del is_abstract

        if not meta.abstract:
            # don't add to model_fields
            meta.fields["pk"] = PKField(exclude=True, name="pk", inherit=False)

        # Handle annotations
        annotations: dict[str, Any] = handle_annotations(bases, base_annotations, attrs)

        for k, _ in meta.managers.items():
            if annotations:
                if k not in annotations:
                    raise ImproperlyConfigured(
                        f"Managers must be type annotated and '{k}' is not annotated. Managers must be annotated with ClassVar."
                    )
                # evaluate annotation which can be a string reference.
                # because we really import ClassVar to check against it is safe to assume a ClassVar is available.
                if isinstance(annotations[k], str):
                    annotations[k] = eval(annotations[k])
                if get_origin(annotations[k]) is not ClassVar:
                    raise ImproperlyConfigured("Managers must be ClassVar type annotated.")

        # Ensure the initialization is only performed for subclasses of EdgyBaseModel
        attrs["__init_annotations__"] = annotations

        new_class = cast(type["Model"], super().__new__(cls, name, bases, attrs, **kwargs))
        meta.model = new_class
        # Ensure initialization is only performed for subclasses of edgy.Model
        # (excluding the edgy.Model class itself).
        if not has_parents:
            return new_class

        # Ensure the model_fields are updated to the latest
        new_class.model_fields = model_fields
        new_class._db_schemas = {}

        # Set the owner of the field, must be done as early as possible
        # don't use meta.fields to not trigger the lazy evaluation
        for value in meta.fields.values():
            value.owner = new_class
        # set the model_class of managers
        for value in meta.managers.values():
            value.owner = new_class

        if getattr(meta, "unique_together", None) is not None:
            if meta.abstract:
                raise ImproperlyConfigured("unique_together cannot be in abstract classes.")
            unique_together = meta.unique_together
            if not isinstance(unique_together, (list, tuple)):
                value_type = type(unique_together).__name__
                raise ImproperlyConfigured(
                    f"unique_together must be a tuple or list. Got {value_type} instead."
                )
            else:
                for value in unique_together:
                    if not isinstance(value, (str, tuple, UniqueConstraint)):
                        raise ValueError(
                            "The values inside the unique_together must be a string, a tuple of strings or an instance of UniqueConstraint."
                        )

        # Handle indexes
        if getattr(meta, "indexes", None) is not None:
            if meta.abstract:
                raise ImproperlyConfigured("indexes cannot be in abstract classes.")
            indexes = meta.indexes
            if not isinstance(indexes, (list, tuple)):
                value_type = type(indexes).__name__
                raise ImproperlyConfigured(
                    f"indexes must be a tuple or list. Got {value_type} instead."
                )
            else:
                for value in indexes:
                    if not isinstance(value, Index):
                        raise ValueError("Meta.indexes must be a list of Index types.")

        # Making sure the tablename is always set if the value is not provided
        if getattr(meta, "tablename", None) is None:
            tablename = f"{name.lower()}s"
            meta.tablename = tablename
        meta.model = new_class
        if skip_registry:
            # don't add automatically to registry. Useful for subclasses which modify the registry itself.
            new_class.model_rebuild(force=True)
            return new_class

        # Now set the registry of models
        if meta.registry is None:
            registry: Union[Registry, None, Literal[False]] = get_model_registry(bases, meta_class)
            meta.registry = registry or None
        if not meta.registry:
            new_class.model_rebuild(force=True)
            return new_class
        new_class.add_to_registry(meta.registry)
        return new_class

    def get_db_schema(cls) -> Union[str, None]:
        """
        Returns a db_schema from registry if any is passed.
        """
        if hasattr(cls, "meta") and getattr(cls.meta, "registry", None):
            return cls.meta.registry.db_schema  # type: ignore
        return None

    def get_db_shema(cls) -> Union[str, None]:
        warnings.warn(
            "'get_db_shema' has been deprecated, use 'get_db_schema' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.get_db_schema()

    def _build_table(cls, metadata: Optional[sqlalchemy.MetaData] = None) -> None:
        try:
            cls._table = cls.build(cls.get_db_schema(), metadata=metadata)
        except AttributeError as exc:
            raise TableBuildError(exc) from exc

    @property
    def table(cls) -> sqlalchemy.Table:
        """
        Making sure the tables on inheritance state, creates the new
        one properly.

        Making sure the following scenarios are met:

        1. If there is a context_db_schema, it will return for those, which means, the `using`
        if being utilised.
        2. If a db_schema in the `registry` is passed, then it will use that as a default.
        3. If none is passed, defaults to the shared schema of the database connected.
        """
        if cls.__is_proxy_model__:
            return cls.__parent__.table  # type: ignore
        if not cls.meta.registry:
            # we cannot set the table without a registry
            # raising is required
            raise AttributeError()
        table = getattr(cls, "_table", None)
        # assert table.name.lower() == cls.meta.tablename, f"{table.name.lower()} != {cls.meta.tablename}"
        # fix assigned table
        if table is None or table.name.lower() != cls.meta.tablename:
            cls._build_table()

        return cast("sqlalchemy.Table", cls._table)

    def __getattr__(cls, name: str) -> Any:
        """
        For controlling inheritance we cannot have manager descriptors, so fake them
        """
        manager = cls.meta.managers.get(name)
        if manager is not None:
            return manager
        return super().__getattr__(name)

    @property
    def pkcolumns(cls) -> Sequence[str]:
        if cls.__dict__.get("_pkcolumns", None) is None:
            build_pkcolumns(cls)
        return cast(Sequence[str], cls._pkcolumns)

    @property
    def pknames(cls) -> Sequence[str]:
        if cls.__dict__.get("_pknames", None) is None:
            build_pknames(cls)
        return cast(Sequence[str], cls._pknames)

    @property
    def signals(cls) -> signals_module.Broadcaster:
        """
        Returns the signals of a class
        """
        warnings.warn(
            "'signals' has been deprecated, use 'meta.signals' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        meta: MetaInfo = cls.meta
        return meta.signals

    def table_schema(
        cls,
        schema: Union[str, None] = None,
        *,
        metadata: Optional[sqlalchemy.MetaData] = None,
        update_cache: bool = False,
    ) -> sqlalchemy.Table:
        """
        Retrieve table for schema (nearly the same as build with scheme argument).
        Cache per class via a primitive LRU cache.
        """
        if cls.__is_proxy_model__:
            return cls.__parent__.table_schema(  # type: ignore
                schema, metadata=metadata, update_cache=update_cache
            )
        if schema is None or (cls.get_db_schema() or "") == schema:
            # sqlalchemy uses "" for empty schema
            table = getattr(cls, "_table", None)
            if update_cache or table is None or table.name.lower() != cls.meta.tablename:
                cls._build_table(metadata=metadata)
            return cls.table
        # remove cache element so the key is reordered
        schema_obj = cls._db_schemas.pop(schema, None)
        if schema_obj is None or update_cache:
            schema_obj = cls.build(schema=schema, metadata=metadata)
        # set element to last
        cls._db_schemas[schema] = schema_obj
        # remove oldest element, when bigger 100
        while len(cls._db_schemas) > 100:
            cls._db_schemas.pop(next(iter(cls._db_schemas)), None)

        return cast("sqlalchemy.Table", schema_obj)

    @property
    def proxy_model(cls: type[Model]) -> type[Model]:
        """
        Returns the proxy_model from the Model when called using the cache.
        """
        if cls.__is_proxy_model__:
            return cls
        if cls.__proxy_model__ is None:
            proxy_model = cls.generate_proxy_model()
            proxy_model.__parent__ = cls
            proxy_model.model_rebuild(force=True)
            cls.__proxy_model__ = proxy_model
        return cls.__proxy_model__

    @property
    def columns(cls) -> sqlalchemy.sql.ColumnCollection:
        return cast("sqlalchemy.sql.ColumnCollection", cls.table.columns)

    @property
    def fields(cls) -> dict[str, BaseFieldType]:
        warnings.warn(
            "'fields' has been deprecated, use 'meta.fields' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        meta: MetaInfo = cls.meta
        return meta.fields



import asyncio
import contextlib
import re
import warnings
from collections import defaultdict
from collections.abc import Mapping, Sequence
from copy import copy as shallow_copy
from functools import cached_property, partial
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)

import sqlalchemy
from loguru import logger
from sqlalchemy import Engine
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.orm import declarative_base as sa_declarative_base

from edgy.core.connection.database import Database, DatabaseURL
from edgy.core.connection.schemas import Schema
from edgy.core.utils.sync import run_sync
from edgy.types import Undefined

from .asgi import ASGIApp, ASGIHelper

if TYPE_CHECKING:
    from edgy.contrib.autoreflection.models import AutoReflectionModel
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.types import BaseModelType


class MetaDataDict(defaultdict[str, sqlalchemy.MetaData]):
    def __init__(self, registry: "Registry") -> None:
        self.registry = registry
        super().__init__(sqlalchemy.MetaData)

    def __getitem__(self, key: Union[str, None]) -> sqlalchemy.MetaData:
        if key not in self.registry.extra and key is not None:
            raise KeyError("Key does not exist")
        return super().__getitem__(key)

    def get(self, key: str, default: Any = None) -> sqlalchemy.MetaData:
        try:
            return self[key]
        except KeyError:
            return default

    def __copy__(self) -> "MetaDataDict":
        _copy = MetaDataDict(registry=self.registry)
        for k, v in self.items():
            _copy[k] = shallow_copy(v)
        return _copy

    copy = __copy__


class MetaDataByUrlDict(dict):
    def __init__(self, registry: "Registry") -> None:
        self.registry = registry
        super().__init__()
        self.process()

    def process(self) -> None:
        self.clear()
        self[str(self.registry.database.url)] = None
        for k, v in self.registry.extra.items():
            self.setdefault(str(v.url), k)

    def __getitem__(self, key: str) -> sqlalchemy.MetaData:
        translation_name = super().__getitem__(key)
        try:
            return self.registry.metadata_by_name[translation_name]
        except KeyError as exc:
            raise Exception("metadata_by_name returned exception") from exc

    def get(self, key: str, default: Any = None) -> sqlalchemy.MetaData:
        try:
            return self[key]
        except KeyError:
            return default

    def __copy__(self) -> "MetaDataByUrlDict":
        return MetaDataByUrlDict(registry=self.registry)

    copy = __copy__


class Registry:
    """
    The command center for the models of Edgy.
    """

    db_schema: Union[str, None] = None
    content_type: Union[type["BaseModelType"], None] = None
    dbs_reflected: set[Union[str, None]]

    def __init__(
        self,
        database: Union[Database, str, DatabaseURL],
        *,
        with_content_type: Union[bool, type["BaseModelType"]] = False,
        schema: Union[str, None] = None,
        extra: Optional[dict[str, Database]] = None,
        **kwargs: Any,
    ) -> None:
        self.db_schema = schema
        extra = extra or {}
        self.database: Database = (
            database if isinstance(database, Database) else Database(database, **kwargs)
        )
        self.models: dict[str, type[BaseModelType]] = {}
        self.reflected: dict[str, type[BaseModelType]] = {}
        self.tenant_models: dict[str, type[BaseModelType]] = {}
        self.pattern_models: dict[str, type[AutoReflectionModel]] = {}
        self.dbs_reflected = set()

        self.schema = Schema(registry=self)
        # when setting a Model or Reflected Model execute the callbacks
        # Note: they are only executed if the Model is not in Registry yet
        self._onetime_callbacks: dict[
            Union[str, None], list[Callable[[type[BaseModelType]], None]]
        ] = defaultdict(list)
        self._callbacks: dict[Union[str, None], list[Callable[[type[BaseModelType]], None]]] = (
            defaultdict(list)
        )

        self.extra: Mapping[str, Database] = {
            k: v if isinstance(v, Database) else Database(v) for k, v in extra.items()
        }
        self.metadata_by_url = MetaDataByUrlDict(registry=self)

        if with_content_type is not False:
            self._set_content_type(with_content_type)

    def __copy__(self) -> "Registry":
        _copy = Registry(self.database)
        _copy.extra = self.extra
        _copy.models = {key: val.copy_edgy_model(_copy) for key, val in self.models.items()}
        _copy.reflected = {key: val.copy_edgy_model(_copy) for key, val in self.reflected.items()}
        _copy.tenant_models = {
            key: val.copy_edgy_model(_copy) for key, val in self.tenant_models.items()
        }
        _copy.pattern_models = {
            key: val.copy_edgy_model(_copy) for key, val in self.pattern_models.items()
        }
        _copy.dbs_reflected = set(self.dbs_reflected)
        if self.content_type is not None:
            try:
                _copy.content_type = self.get_model("ContentType")
            except LookupError:
                _copy.content_type = self.content_type
            # init callbacks
            _copy._set_content_type(_copy.content_type)
        return _copy

    def _set_content_type(
        self, with_content_type: Union[Literal[True], type["BaseModelType"]]
    ) -> None:
        from edgy.contrib.contenttypes.fields import BaseContentTypeFieldField, ContentTypeField
        from edgy.contrib.contenttypes.models import ContentType
        from edgy.core.db.models.metaclasses import MetaInfo
        from edgy.core.db.relationships.related_field import RelatedField
        from edgy.core.utils.models import create_edgy_model

        if with_content_type is True:
            with_content_type = ContentType

        real_content_type: type[BaseModelType] = with_content_type

        if real_content_type.meta.abstract:
            meta_args = {
                "tablename": "contenttypes",
                "registry": self,
            }

            new_meta: MetaInfo = MetaInfo(None, **meta_args)
            # model adds itself to registry and executes callbacks
            real_content_type = create_edgy_model(
                "ContentType",
                with_content_type.__module__,
                __metadata__=new_meta,
                __bases__=(with_content_type,),
            )
        elif real_content_type.meta.registry is None:
            real_content_type.add_to_registry(self, "ContentType")
        self.content_type = real_content_type

        def callback(model_class: type["BaseModelType"]) -> None:
            # they are not updated, despite this shouldn't happen anyway
            if issubclass(model_class, ContentType):
                return
            # skip if is explicit set
            for field in model_class.meta.fields.values():
                if isinstance(field, BaseContentTypeFieldField):
                    return
            # e.g. exclude field
            if "content_type" not in model_class.meta.fields:
                related_name = f"reverse_{model_class.__name__.lower()}"
                assert (
                    related_name not in real_content_type.meta.fields
                ), f"duplicate model name: {model_class.__name__}"
                model_class.meta.fields["content_type"] = cast(
                    "BaseFieldType",
                    ContentTypeField(
                        name="content_type",
                        owner=model_class,
                        to=real_content_type,
                        no_constraint=real_content_type.no_constraint,
                    ),
                )
                real_content_type.meta.fields[related_name] = RelatedField(
                    name=related_name,
                    foreign_key_name="content_type",
                    related_from=model_class,
                    owner=real_content_type,
                )

        self.register_callback(None, callback, one_time=False)

    @property
    def metadata_by_name(self) -> MetaDataDict:
        if getattr(self, "_metadata_by_name", None) is None:
            self._metadata_by_name = MetaDataDict(registry=self)
        return self._metadata_by_name

    @metadata_by_name.setter
    def metadata_by_name(self, value: MetaDataDict) -> None:
        metadata_dict = self.metadata_by_name
        metadata_dict.clear()
        for k, v in value.items():
            metadata_dict[k] = v
        self.metadata_by_url.process()

    @property
    def metadata(self) -> sqlalchemy.MetaData:
        warnings.warn(
            "metadata is deprecated use metadata_by_name or metadata_by_url instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.metadata_by_name[None]

    def get_model(self, model_name: str) -> type["BaseModelType"]:
        if model_name in self.models:
            return self.models[model_name]
        elif model_name in self.reflected:
            return self.reflected[model_name]
        elif model_name in self.tenant_models:
            return self.tenant_models[model_name]
        else:
            raise LookupError(f"Registry doesn't have a {model_name} model.") from None

    def refresh_metadata(
        self,
        *,
        update_only: bool = False,
        multi_schema: Union[bool, re.Pattern, str] = False,
        ignore_schema_pattern: Union[None, "re.Pattern", str] = "information_schema",
    ) -> None:
        if not update_only:
            for val in self.metadata_by_name.values():
                val.clear()
        maindatabase_url = str(self.database.url)
        if multi_schema is not False:
            schemes_tree: dict[str, tuple[Optional[str], list[str]]] = {
                v[0]: (key, v[2])
                for key, v in run_sync(self.schema.get_schemes_tree(no_reflect=True)).items()
            }
        else:
            schemes_tree = {
                maindatabase_url: (None, [self.db_schema]),
                **{str(v.url): (k, [None]) for k, v in self.extra.items()},
            }

        if isinstance(multi_schema, str):
            multi_schema = re.compile(multi_schema)
        if isinstance(ignore_schema_pattern, str):
            ignore_schema_pattern = re.compile(ignore_schema_pattern)
        for model_class in self.models.values():
            if not update_only:
                model_class._table = None
                model_class._db_schemas = {}
            url = str(model_class.database.url)
            if url in schemes_tree:
                extra_key, schemes = schemes_tree[url]
                for schema in schemes:
                    if multi_schema is not False:
                        if multi_schema is not True and multi_schema.match(schema) is None:
                            continue
                        if (
                            ignore_schema_pattern is not None
                            and ignore_schema_pattern.match(schema) is not None
                        ):
                            continue
                        if not getattr(model_class.meta, "is_tenant", False):
                            if (
                                model_class.__using_schema__ is Undefined
                                or model_class.__using_schema__ is None
                            ):
                                if schema != "":
                                    continue
                            elif model_class.__using_schema__ != schema:
                                continue
                    model_class.table_schema(schema=schema, metadata=self.metadata_by_url[url])

        # don't initialize to keep the metadata clean
        if not update_only:
            for model_class in self.reflected.values():
                model_class._table = None
                model_class._db_schemas = {}

    def register_callback(
        self,
        name_or_class: Union[type["BaseModelType"], str, None],
        callback: Callable[[type["BaseModelType"]], None],
        one_time: Optional[bool] = None,
    ) -> None:
        if one_time is None:
            # True for model specific callbacks, False for general callbacks
            one_time = name_or_class is not None
        called: bool = False
        if name_or_class is None:
            for model in self.models.values():
                callback(model)
                called = True
            for model in self.reflected.values():
                callback(model)
                called = True
            for name, model in self.tenant_models.items():
                # for tenant only models
                if name not in self.models:
                    callback(model)
                    called = True
        elif not isinstance(name_or_class, str):
            callback(name_or_class)
            called = True
        else:
            model_class = None
            with contextlib.suppress(LookupError):
                model_class = self.get_model(name_or_class)
            if model_class is not None:
                callback(model_class)
                called = True
        if name_or_class is not None and not isinstance(name_or_class, str):
            name_or_class = name_or_class.__name__
        if called and one_time:
            return
        if one_time:
            self._onetime_callbacks[name_or_class].append(callback)
        else:
            self._callbacks[name_or_class].append(callback)

    def execute_model_callbacks(self, model_class: type["BaseModelType"]) -> None:
        name = model_class.__name__
        callbacks = self._onetime_callbacks.get(name)
        while callbacks:
            callbacks.pop()(model_class)

        callbacks = self._onetime_callbacks.get(None)
        while callbacks:
            callbacks.pop()(model_class)

        callbacks = self._callbacks.get(name)
        if callbacks:
            for callback in callbacks:
                callback(model_class)

        callbacks = self._callbacks.get(None)
        if callbacks:
            for callback in callbacks:
                callback(model_class)

    @cached_property
    def declarative_base(self) -> Any:
        if self.db_schema:
            metadata = sqlalchemy.MetaData(schema=self.db_schema)
        else:
            metadata = sqlalchemy.MetaData()
        return sa_declarative_base(metadata=metadata)

    @property
    def engine(self) -> AsyncEngine:
        assert self.database.is_connected and self.database.engine, "database not initialized"
        return self.database.engine

    @property
    def sync_engine(self) -> Engine:
        return self.engine.sync_engine

    def init_models(
        self, *, init_column_mappers: bool = True, init_class_attrs: bool = True
    ) -> None:
        """
        Initializes lazy parts of models meta. Normally not needed to call.
        """
        for model_class in self.models.values():
            model_class.meta.full_init(
                init_column_mappers=init_column_mappers, init_class_attrs=init_class_attrs
            )

        for model_class in self.reflected.values():
            model_class.meta.full_init(
                init_column_mappers=init_column_mappers, init_class_attrs=init_class_attrs
            )

    def invalidate_models(self, *, clear_class_attrs: bool = True) -> None:
        """
        Invalidate all lazy parts of meta. They will automatically re-initialized on access.
        """
        for model_class in self.models.values():
            model_class.meta.invalidate(clear_class_attrs=clear_class_attrs)
        for model_class in self.reflected.values():
            model_class.meta.invalidate(clear_class_attrs=clear_class_attrs)

    def get_tablenames(self) -> set[str]:
        return_set = set()
        for model_class in self.models.values():
            return_set.add(model_class.meta.tablename)
        for model_class in self.reflected.values():
            return_set.add(model_class.meta.tablename)
        return return_set

    async def _connect_and_init(self, name: Union[str, None], database: "Database") -> None:
        from edgy.core.db.models.metaclasses import MetaInfo

        await database.connect()
        if not self.pattern_models or name in self.dbs_reflected:
            return
        schemes = set()
        for pattern_model in self.pattern_models.values():
            if name not in pattern_model.meta.databases:
                continue
            schemes.update(pattern_model.meta.schemes)
        tmp_metadata = sqlalchemy.MetaData()
        for schema in schemes:
            await database.run_sync(tmp_metadata.reflect, schema=schema)
        try:
            for table in tmp_metadata.tables.values():
                for pattern_model in self.pattern_models.values():
                    if name not in pattern_model.meta.databases or table.schema not in schemes:
                        continue
                    assert pattern_model.meta.model is pattern_model
                    # table.key would contain the schema name
                    if not pattern_model.meta.include_pattern.match(table.name) or (
                        pattern_model.meta.exclude_pattern
                        and pattern_model.meta.exclude_pattern.match(table.name)
                    ):
                        continue
                    if pattern_model.fields_not_supported_by_table(table):
                        continue
                    new_name = pattern_model.meta.template(table)
                    old_model: Optional[type[BaseModelType]] = None
                    with contextlib.suppress(LookupError):
                        old_model = self.get_model(new_name)
                    if old_model is not None:
                        raise Exception(
                            f"Conflicting model: {old_model.__name__} with pattern model: {pattern_model.__name__}"
                        )
                    concrete_reflect_model = pattern_model.copy_edgy_model(
                        name=new_name, meta_info_class=MetaInfo
                    )
                    concrete_reflect_model.meta.tablename = table.name
                    concrete_reflect_model.__using_schema__ = table.schema
                    concrete_reflect_model.add_to_registry(self, database=database)

            self.dbs_reflected.add(name)
        except BaseException as exc:
            await database.disconnect()
            raise exc

    async def __aenter__(self) -> "Registry":
        dbs: list[tuple[Union[str, None], Database]] = [(None, self.database)]
        for name, db in self.extra.items():
            dbs.append((name, db))
        ops = [self._connect_and_init(name, db) for name, db in dbs]
        results: list[Union[BaseException, bool]] = await asyncio.gather(
            *ops, return_exceptions=True
        )
        if any(isinstance(x, BaseException) for x in results):
            ops2 = []
            for num, value in enumerate(results):
                if not isinstance(value, BaseException):
                    ops2.append(dbs[num][1].disconnect())
                else:
                    logger.opt(exception=value).error("Failed to connect database.")
            await asyncio.gather(*ops2)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]] = None,
        exc_value: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> None:
        ops = [self.database.disconnect()]
        for value in self.extra.values():
            ops.append(value.disconnect())
        await asyncio.gather(*ops)

    @overload
    def asgi(
        self,
        app: None,
        handle_lifespan: bool = False,
    ) -> Callable[[ASGIApp], ASGIHelper]: ...

    @overload
    def asgi(
        self,
        app: ASGIApp,
        handle_lifespan: bool = False,
    ) -> ASGIHelper: ...

    def asgi(
        self,
        app: Optional[ASGIApp] = None,
        handle_lifespan: bool = False,
    ) -> Union[ASGIHelper, Callable[[ASGIApp], ASGIHelper]]:
        """Return wrapper for asgi integration."""
        if app is not None:
            return ASGIHelper(app=app, registry=self, handle_lifespan=handle_lifespan)
        return partial(ASGIHelper, registry=self, handle_lifespan=handle_lifespan)

    async def create_all(
        self, refresh_metadata: bool = True, databases: Sequence[Union[str, None]] = (None,)
    ) -> None:
        # otherwise old references to non-existing tables, fks can lurk around
        if refresh_metadata:
            self.refresh_metadata(multi_schema=True)
        if self.db_schema:
            await self.schema.create_schema(
                self.db_schema, True, True, update_cache=True, databases=databases
            )
        else:
            # fallback when no schemes are in use. Because not all dbs support schemes
            # we cannot just use a scheme = ""
            for database in databases:
                db = self.database if database is None else self.extra[database]
                # don't warn here about inperformance
                async with db as db:
                    with db.force_rollback(False):
                        await db.create_all(self.metadata_by_name[database])

    async def drop_all(self, databases: Sequence[Union[str, None]] = (None,)) -> None:
        if self.db_schema:
            await self.schema.drop_schema(
                self.db_schema, cascade=True, if_exists=True, databases=databases
            )
        else:
            # fallback when no schemes are in use. Because not all dbs support schemes
            # we cannot just use a scheme = ""
            for database_name in databases:
                db = self.database if database_name is None else self.extra[database_name]
                # don't warn here about inperformance
                async with db as db:
                    with db.force_rollback(False):
                        await db.drop_all(self.metadata_by_name[database_name])



from functools import cached_property
from typing import TYPE_CHECKING, Any, Union

from edgy.conf import settings
from edgy.conf.module_import import import_string
from edgy.exceptions import InvalidStorageError

if TYPE_CHECKING:
    from edgy.core.files.storage.base import Storage


class StorageHandler:
    """
    The handler for the available storages
    of the system.
    """

    def __init__(self, backends: Union[dict[str, Any], None] = None) -> None:
        self._backends = backends
        self._storages: dict[str, Storage] = {}

    @cached_property
    def backends(self) -> dict[str, Any]:
        if self._backends is None:
            self._backends = settings.storages.copy()
        return self._backends

    def __getitem__(self, alias: str) -> "Storage":
        """
        Get the storage instance associated with the given alias.

        Args:
            alias (str): The alias of the storage.

        Returns:
            storage: The storage instance.

        Raises:
            InvalidStorageError: If the alias is not found in settings.STORAGES.
        """
        storage = self._storages.get(alias, None)
        if storage is not None:
            return storage

        params = self.backends.get(alias, None)
        if params is None:
            raise InvalidStorageError(f"Could not find config for '{alias}' in settings.STORAGES.")

        storage = self.create_storage(params)
        storage.name = alias
        self._storages[alias] = storage
        return storage

    def create_storage(self, params: dict[str, Any]) -> "Storage":
        """
        Create a storage instance based on the provided parameters.

        Args:
            params (dict): The parameters for configuring the storage.

        Returns:
            storage: The storage instance.

        Raises:
            InvalidStorageError: If the backend specified in params cannot be imported.
        """
        backend = params.pop("backend")
        options = params.pop("options", {})

        try:
            storage_cls: type[Storage] = import_string(backend)
        except ImportError as e:
            raise InvalidStorageError(f"Could not find backend {backend!r}: {e}") from e
        return storage_cls(**options)



import datetime
from collections import OrderedDict
from typing import Any

import pydantic

import edgy
from edgy import Registry
from edgy.core.terminal import OutputColour, Print

printer = Print()

defaults = OrderedDict()

defaults.update(
    {
        "datetime": datetime.datetime,
        "timedelta": datetime.timedelta,
        "BaseModel": pydantic.BaseModel,
        "ConfigDict": pydantic.ConfigDict,
        "settings": edgy.settings,
        "Model": edgy.Model,
        "ReflectModel": edgy.ReflectModel,
    }
)


def welcome_message(app: Any) -> None:
    """Displays the welcome message for the user"""
    now = datetime.datetime.now().strftime("%b %d %Y, %H:%M:%S")
    edgy_info_date = f"Edgy {edgy.__version__} (interactive shell, {now})"
    info = "Interactive shell that imports the application models and some python defaults."

    application_text = printer.message("Application: ", colour=OutputColour.CYAN3)
    application_name = printer.message(app.__class__.__name__, colour=OutputColour.GREEN3)
    application = f"{application_text}{application_name}"

    printer.write_plain(edgy_info_date, colour=OutputColour.CYAN3)
    printer.write_plain(info, colour=OutputColour.CYAN3)
    printer.write_plain(application)


def import_objects(app: Any, registry: Registry) -> dict[Any, Any]:
    """
    Imports all the needed objects needed for the shell.
    """
    imported_objects = {}
    import_statement = "from {module_path} import {model}"
    welcome_message(app)
    printer.write_success(79 * "-", colour=OutputColour.CYAN3)

    def import_defaults() -> None:
        for name, module in defaults.items():
            directive = import_statement.format(module_path=module.__module__, model=name)
            printer.write_success(directive, colour=OutputColour.CYAN3)
            imported_objects[name] = module

    def _import_objects(lookup_dict: dict[Any, Any]) -> None:
        for _, model in sorted(lookup_dict.items()):
            directive = import_statement.format(module_path=model.__module__, model=model.__name__)
            printer.write_success(directive, colour=OutputColour.CYAN3)
            imported_objects[model.__name__] = model

    def import_models() -> None:
        if not registry.models:
            return

        printer.write_success("Models".center(79, "-"), colour=OutputColour.CYAN3)
        _import_objects(registry.models)

    def import_reflected_models() -> None:
        if not registry.reflected:
            return

        printer.write_success("Reflected models".center(79, "-"), colour=OutputColour.CYAN3)
        _import_objects(registry.reflected)

    import_defaults()
    import_models()
    import_reflected_models()

    return imported_objects



import contextlib
import inspect
import warnings
from collections.abc import Sequence
from functools import partial
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, cast

import sqlalchemy
from pydantic import BaseModel

from edgy.core.db.context_vars import (
    CURRENT_INSTANCE,
    EXPLICIT_SPECIFIED_VALUES,
    MODEL_GETATTR_BEHAVIOR,
    get_schema,
)
from edgy.core.db.datastructures import Index, UniqueConstraint
from edgy.core.db.fields.many_to_many import BaseManyToManyForeignKeyField
from edgy.core.db.models.metaclasses import MetaInfo
from edgy.core.db.models.utils import build_pkcolumns, build_pknames
from edgy.core.db.relationships.related_field import RelatedField
from edgy.core.utils.db import check_db_connection
from edgy.exceptions import ForeignKeyBadConfigured, ObjectNotFound
from edgy.types import Undefined

if TYPE_CHECKING:
    from databasez.core.transaction import Transaction

    from edgy.core.connection.database import Database
    from edgy.core.connection.registry import Registry
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.model import Model
    from edgy.core.db.models.types import BaseModelType


_empty = cast(set[str], frozenset())


def _set_related_field(
    target: type["BaseModelType"],
    *,
    foreign_key_name: str,
    related_name: str,
    source: type["BaseModelType"],
    replace_related_field: bool,
) -> None:
    if not replace_related_field and related_name in target.meta.fields:
        # is already correctly set, required for migrate of model_apps with registry set
        related_field = target.meta.fields[related_name]
        if (
            related_field.related_from is source
            and related_field.foreign_key_name == foreign_key_name
        ):
            return
        raise ForeignKeyBadConfigured(
            f"Multiple related_name with the same value '{related_name}' found to the same target. Related names must be different."
        )

    related_field = RelatedField(
        foreign_key_name=foreign_key_name,
        name=related_name,
        owner=target,
        related_from=source,
    )

    # Set the related name
    target.meta.fields[related_name] = related_field


def _set_related_name_for_foreign_keys(
    meta: "MetaInfo", model_class: type["BaseModelType"], replace_related_field: bool = False
) -> None:
    """
    Sets the related name for the foreign keys.
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """
    if not meta.foreign_key_fields:
        return

    for name in meta.foreign_key_fields:
        foreign_key = meta.fields[name]
        related_name = getattr(foreign_key, "related_name", None)
        if related_name is False:
            # skip related_field
            continue

        if not related_name:
            if foreign_key.unique:
                related_name = f"{model_class.__name__.lower()}"
            else:
                related_name = f"{model_class.__name__.lower()}s_set"

        foreign_key.related_name = related_name
        foreign_key.reverse_name = related_name

        related_field_fn = partial(
            _set_related_field,
            source=model_class,
            foreign_key_name=name,
            related_name=related_name,
            replace_related_field=replace_related_field,
        )
        registry: Registry = foreign_key.target_registry
        with contextlib.suppress(Exception):
            registry = cast("Registry", foreign_key.target.registry)
        registry.register_callback(foreign_key.to, related_field_fn, one_time=True)


class DatabaseMixin:
    @classmethod
    def add_to_registry(
        cls: type["BaseModelType"],
        registry: "Registry",
        name: str = "",
        database: Union[bool, "Database", Literal["keep"]] = "keep",
    ) -> None:
        # when called if registry is not set
        cls.meta.registry = registry
        if database is True:
            cls.database = registry.database
        elif database is not False:
            if database == "keep":
                if getattr(cls, "database", None) is None:
                    cls.database = registry.database

            else:
                cls.database = database
        meta = cls.meta
        if name:
            cls.__name__ = name

        # Making sure it does not generate models if abstract or a proxy
        if not meta.abstract and not cls.__is_proxy_model__:
            if getattr(cls, "__reflected__", False):
                registry.reflected[cls.__name__] = cls
            else:
                registry.models[cls.__name__] = cls
            # after registrating the own model
            for value in list(meta.fields.values()):
                if isinstance(value, BaseManyToManyForeignKeyField):
                    m2m_registry: Registry = value.target_registry
                    with contextlib.suppress(Exception):
                        m2m_registry = cast("Registry", value.target.registry)

                    def create_through_model(x: Any, field: "BaseFieldType" = value) -> None:
                        # we capture with field = ... the variable
                        field.create_through_model()

                    m2m_registry.register_callback(value.to, create_through_model, one_time=True)
            # Sets the foreign key fields
            if meta.foreign_key_fields:
                _set_related_name_for_foreign_keys(meta, cls)
            registry.execute_model_callbacks(cls)

        # finalize
        cls.model_rebuild(force=True)

    def get_active_instance_schema(
        self, check_schema: bool = True, check_tenant: bool = True
    ) -> Union[str, None]:
        if self.__using_schema__ is not Undefined:
            return cast(Union[str, None], self.__using_schema__)
        return self.__class__.get_active_class_schema(
            check_schema=check_schema, check_tenant=check_tenant
        )

    @classmethod
    def get_active_class_schema(cls, check_schema: bool = True, check_tenant: bool = True) -> str:
        if cls.__using_schema__ is not Undefined:
            return cast(Union[str, None], cls.__using_schema__)
        if check_schema:
            schema = get_schema(check_tenant=check_tenant)
            if schema is not None:
                return schema
        db_schema: Optional[str] = cls.get_db_schema()
        # sometime "" is ok, sometimes not, sqlalchemy logic
        return db_schema or None

    @classmethod
    def copy_edgy_model(
        cls: type["Model"], registry: Optional["Registry"] = None, name: str = "", **kwargs: Any
    ) -> type["Model"]:
        """Copy the model class and optionally add it to another registry."""
        # removes private pydantic stuff, except the prefixed ones
        attrs = {
            key: val
            for key, val in cls.__dict__.items()
            if key not in BaseModel.__dict__ or key.startswith("__")
        }
        attrs.pop("meta", None)
        # managers and fields are gone, we have to readd them with the correct data
        attrs.update(cls.meta.fields)
        attrs.update(cls.meta.managers)
        _copy = cast(
            type["Model"],
            type(cls.__name__, cls.__bases__, attrs, skip_registry=True, **kwargs),
        )
        _copy.meta.model = _copy
        if name:
            _copy.__name__ = name
        if registry is not None:
            _copy.add_to_registry(registry)
        return _copy

    @property
    def table(self) -> sqlalchemy.Table:
        if getattr(self, "_table", None) is None:
            schema = self.get_active_instance_schema()
            return cast(
                "sqlalchemy.Table",
                self.__class__.table_schema(schema),
            )
        return self._table

    @table.setter
    def table(self, value: sqlalchemy.Table) -> None:
        with contextlib.suppress(AttributeError):
            del self._pknames
        with contextlib.suppress(AttributeError):
            del self._pkcolumns
        self._table = value

    @property
    def pkcolumns(self) -> Sequence[str]:
        if self.__dict__.get("_pkcolumns", None) is None:
            if self.__dict__.get("_table", None) is None:
                self._pkcolumns: Sequence[str] = cast(Sequence[str], self.__class__.pkcolumns)
            else:
                build_pkcolumns(self)
        return self._pkcolumns

    @property
    def pknames(self) -> Sequence[str]:
        if self.__dict__.get("_pknames", None) is None:
            if self.__dict__.get("_table", None) is None:
                self._pknames: Sequence[str] = cast(Sequence[str], self.__class__.pknames)
            else:
                build_pknames(self)
        return self._pknames

    def get_columns_for_name(self: "Model", name: str) -> Sequence["sqlalchemy.Column"]:
        table = self.table
        meta = self.meta
        if name in meta.field_to_columns:
            return meta.field_to_columns[name]
        elif name in table.columns:
            return (table.columns[name],)
        else:
            return cast(Sequence["sqlalchemy.Column"], _empty)

    def identifying_clauses(self, prefix: str = "") -> list[Any]:
        # works only if the class of the model is the main class of the queryset
        # TODO: implement prefix handling and return generic column without table attached
        if prefix:
            raise NotImplementedError()
        clauses: list[Any] = []
        for field_name in self.identifying_db_fields:
            field = self.meta.fields.get(field_name)
            if field is not None:
                for column_name, value in field.clean(
                    field_name, self.__dict__[field_name]
                ).items():
                    clauses.append(getattr(self.table.columns, column_name) == value)
            else:
                clauses.append(
                    getattr(self.table.columns, field_name) == self.__dict__[field_name]
                )
        return clauses

    async def _update(self: "Model", **kwargs: Any) -> Any:
        """
        Update operation of the database fields.
        """
        await self.meta.signals.pre_update.send_async(self.__class__, instance=self)
        column_values = self.extract_column_values(
            extracted_values=kwargs,
            is_partial=True,
            is_update=True,
            phase="prepare_update",
            instance=self,
            model_instance=self,
        )
        # empty updates shouldn't cause an error. E.g. only model references are updated
        clauses = self.identifying_clauses()
        token = CURRENT_INSTANCE.set(self)
        try:
            if column_values and clauses:
                check_db_connection(self.database, stacklevel=4)
                async with self.database as database, database.transaction():
                    # can update column_values
                    column_values.update(
                        await self.execute_pre_save_hooks(
                            column_values, kwargs, force_insert=False
                        )
                    )
                    expression = self.table.update().values(**column_values).where(*clauses)
                    await database.execute(expression)

                # Update the model instance.
                new_kwargs = self.transform_input(
                    column_values, phase="post_update", instance=self
                )
                self.__dict__.update(new_kwargs)

            # updates aren't required to change the db, they can also just affect the meta fields
            await self.execute_post_save_hooks(
                cast(Sequence[str], kwargs.keys()), force_insert=False
            )

        finally:
            CURRENT_INSTANCE.reset(token)
        if column_values or kwargs:
            # Ensure on access refresh the results is active
            self._loaded_or_deleted = False
        await self.meta.signals.post_update.send_async(self.__class__, instance=self)

    async def update(self: "Model", **kwargs: Any) -> "Model":
        token = EXPLICIT_SPECIFIED_VALUES.set(set(kwargs.keys()))
        try:
            await self._update(**kwargs)
        finally:
            EXPLICIT_SPECIFIED_VALUES.reset(token)
        return self

    async def delete(
        self, skip_post_delete_hooks: bool = False, remove_referenced_call: bool = False
    ) -> None:
        """Delete operation from the database"""
        await self.meta.signals.pre_delete.send_async(self.__class__, instance=self)
        # get values before deleting
        field_values: dict[str, Any] = {}
        if not skip_post_delete_hooks and self.meta.post_delete_fields:
            token = MODEL_GETATTR_BEHAVIOR.set("coro")
            try:
                for field_name in self.meta.post_delete_fields:
                    try:
                        field_value = getattr(self, field_name)
                    except AttributeError:
                        continue
                    if inspect.isawaitable(field_value):
                        try:
                            field_value = await field_value
                        except AttributeError:
                            continue
                    field_values[field_name] = field_value
            finally:
                MODEL_GETATTR_BEHAVIOR.reset(token)
        clauses = self.identifying_clauses()
        if clauses:
            expression = self.table.delete().where(*clauses)
            check_db_connection(self.database)
            async with self.database as database:
                await database.execute(expression)
        # we cannot load anymore
        self._loaded_or_deleted = True
        # now cleanup with the saved values
        for field_name, value in field_values.items():
            field = self.meta.fields[field_name]
            await field.post_delete_callback(value, instance=self)

        await self.meta.signals.post_delete.send_async(self.__class__, instance=self)

    async def load(self, only_needed: bool = False) -> None:
        if only_needed and self._loaded_or_deleted:
            return
        row = None
        clauses = self.identifying_clauses()
        if clauses:
            # Build the select expression.
            expression = self.table.select().where(*clauses)

            # Perform the fetch.
            check_db_connection(self.database)
            async with self.database as database:
                row = await database.fetch_one(expression)
        # check if is in system
        if row is None:
            raise ObjectNotFound("row does not exist anymore")
        # Update the instance.
        self.__dict__.update(self.transform_input(dict(row._mapping), phase="load", instance=self))
        self._loaded_or_deleted = True

    async def _insert(self: "Model", **kwargs: Any) -> "Model":
        """
        Performs the save instruction.
        """
        column_values: dict[str, Any] = self.extract_column_values(
            extracted_values=kwargs,
            is_partial=False,
            is_update=False,
            phase="prepare_insert",
            instance=self,
            model_instance=self,
        )
        check_db_connection(self.database, stacklevel=4)
        token = CURRENT_INSTANCE.set(self)
        try:
            async with self.database as database, database.transaction():
                # can update column_values
                column_values.update(
                    await self.execute_pre_save_hooks(column_values, kwargs, force_insert=True)
                )
                expression = self.table.insert().values(**column_values)
                autoincrement_value = await database.execute(expression)
            # sqlalchemy supports only one autoincrement column
            if autoincrement_value:
                column = self.table.autoincrement_column
                if column is not None and hasattr(autoincrement_value, "_mapping"):
                    autoincrement_value = autoincrement_value._mapping[column.key]
                # can be explicit set, which causes an invalid value returned
                if column is not None and column.key not in column_values:
                    column_values[column.key] = autoincrement_value

            new_kwargs = self.transform_input(column_values, phase="post_insert", instance=self)
            self.__dict__.update(new_kwargs)

            if self.meta.post_save_fields:
                await self.execute_post_save_hooks(
                    cast(Sequence[str], kwargs.keys()), force_insert=True
                )
        finally:
            CURRENT_INSTANCE.reset(token)
        # Ensure on access refresh the results is active
        self._loaded_or_deleted = False

        return self

    async def save(
        self: "Model",
        force_insert: bool = False,
        values: Union[dict[str, Any], set[str], None] = None,
        force_save: Optional[bool] = None,
    ) -> "Model":
        """
        Performs a save of a given model instance.
        When creating a user it will make sure it can update existing or
        create a new one.
        """
        if force_save is not None:
            warnings.warn(
                "'force_save' is deprecated in favor of 'force_insert'",
                DeprecationWarning,
                stacklevel=2,
            )
            force_insert = force_save

        await self.meta.signals.pre_save.send_async(self.__class__, instance=self)

        extracted_fields = self.extract_db_fields()
        if values is None:
            explicit_values: set[str] = set()
        elif isinstance(values, set):
            # special mode for marking values as explicit values
            explicit_values = set(values)
            values = None
        else:
            explicit_values = set(values.keys())

        token = MODEL_GETATTR_BEHAVIOR.set("coro")
        try:
            for pkcolumn in self.__class__.pkcolumns:
                # should trigger load in case of identifying_db_fields
                value = getattr(self, pkcolumn, None)
                if inspect.isawaitable(value):
                    value = await value
                if value is None and self.table.columns[pkcolumn].autoincrement:
                    extracted_fields.pop(pkcolumn, None)
                    force_insert = True
                field = self.meta.fields.get(pkcolumn)
                # this is an IntegerField with primary_key set
                if field is not None and getattr(field, "increment_on_save", 0) != 0:
                    # we create a new revision.
                    force_insert = True
                    # Note: we definitely want this because it is easy for forget a force_insert
        finally:
            MODEL_GETATTR_BEHAVIOR.reset(token)

        token2 = EXPLICIT_SPECIFIED_VALUES.set(explicit_values)
        try:
            if force_insert:
                if values:
                    extracted_fields.update(values)
                # force save must ensure a complete mapping
                await self._insert(**extracted_fields)
            else:
                await self._update(**(extracted_fields if values is None else values))
        finally:
            EXPLICIT_SPECIFIED_VALUES.reset(token2)
        await self.meta.signals.post_save.send_async(self.__class__, instance=self)
        return self

    @classmethod
    def build(
        cls, schema: Optional[str] = None, metadata: Optional[sqlalchemy.MetaData] = None
    ) -> sqlalchemy.Table:
        """
        Builds the SQLAlchemy table representation from the loaded fields.
        """
        tablename: str = cls.meta.tablename
        registry = cls.meta.registry
        assert registry, "registry is not set"
        if metadata is None:
            metadata = registry.metadata_by_url[str(cls.database.url)]
        schemes: list[str] = []
        if schema:
            schemes.append(schema)
        if cls.__using_schema__ is not Undefined:
            schemes.append(cls.__using_schema__)
        db_schema = cls.get_db_schema() or ""
        schemes.append(db_schema)

        unique_together = cls.meta.unique_together
        index_constraints = cls.meta.indexes

        columns: list[sqlalchemy.Column] = []
        global_constraints: list[Any] = []
        for name, field in cls.meta.fields.items():
            current_columns = field.get_columns(name)
            columns.extend(current_columns)
            global_constraints.extend(field.get_global_constraints(name, current_columns, schemes))

        # Handle the uniqueness together
        uniques = []
        for unique_index in unique_together or []:
            unique_constraint = cls._get_unique_constraints(unique_index)
            uniques.append(unique_constraint)

        # Handle the indexes
        indexes = []
        for index_c in index_constraints or []:
            index = cls._get_indexes(index_c)
            indexes.append(index)
        return sqlalchemy.Table(
            tablename,
            metadata,
            *columns,
            *uniques,
            *indexes,
            *global_constraints,
            extend_existing=True,
            schema=schema
            if schema
            else cls.get_active_class_schema(check_schema=False, check_tenant=False),
        )

    @classmethod
    def _get_unique_constraints(
        cls, columns: Union[Sequence, str, sqlalchemy.UniqueConstraint]
    ) -> Optional[sqlalchemy.UniqueConstraint]:
        """
        Returns the unique constraints for the model.

        The columns must be a a list, tuple of strings or a UniqueConstraint object.

        :return: Model UniqueConstraint.
        """
        if isinstance(columns, str):
            return sqlalchemy.UniqueConstraint(columns)
        elif isinstance(columns, UniqueConstraint):
            return sqlalchemy.UniqueConstraint(*columns.fields, name=columns.name)
        return sqlalchemy.UniqueConstraint(*columns)

    @classmethod
    def _get_indexes(cls, index: Index) -> Optional[sqlalchemy.Index]:
        """
        Creates the index based on the Index fields
        """
        return sqlalchemy.Index(index.name, *index.fields)

    def transaction(self, *, force_rollback: bool = False, **kwargs: Any) -> "Transaction":
        """Return database transaction for the assigned database"""
        return cast(
            "Transaction", self.database.transaction(force_rollback=force_rollback, **kwargs)
        )



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, cast

from pydantic import BaseModel, ConfigDict
from sqlalchemy.exc import IntegrityError

from edgy.core.db.fields.base import RelationshipField
from edgy.exceptions import ObjectNotFound, RelationshipIncompatible, RelationshipNotFound
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:
    from edgy import QuerySet
    from edgy.core.db.models.types import BaseModelType


class ManyRelation(ManyRelationProtocol):
    """
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(
        self,
        *,
        from_foreign_key: str,
        to_foreign_key: str,
        to: type["BaseModelType"],
        through: type["BaseModelType"],
        reverse: bool = False,
        embed_through: Union[Literal[False], str] = "",
        refs: Any = (),
        instance: Optional[Union["BaseModelType"]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.through = through
        self.to = to
        self.instance = instance
        self.reverse = reverse
        self.from_foreign_key = from_foreign_key
        self.to_foreign_key = to_foreign_key
        self.embed_through = embed_through
        self.refs: Sequence[BaseModelType] = []
        if not isinstance(refs, Sequence):
            refs = [refs]
        self.stage(*refs)

    def get_queryset(self) -> "QuerySet":
        # we need to check tenant every request
        queryset = self.through.meta.managers["query_related"].get_queryset()
        assert self.instance, "instance not initialized"
        fk = self.through.meta.fields[self.from_foreign_key]
        query = {}
        for related_name in fk.related_columns:
            query[related_name] = getattr(self.instance, related_name)
        queryset = queryset.filter(**{self.from_foreign_key: query})
        # now set embed_parent
        queryset.embed_parent = (self.to_foreign_key, self.embed_through or "")
        if self.embed_through:
            queryset.embed_parent_filters = queryset.embed_parent
        return queryset

    async def save_related(self) -> None:
        # TODO: improve performance
        fk = self.through.meta.fields[self.from_foreign_key]
        while self.refs:
            ref = self.refs.pop()
            ref.__dict__.update(fk.clean(fk.name, self.instance))
            await self.add(ref)

    def __getattr__(self, item: Any) -> Any:
        """
        Gets the attribute from the queryset and if it does not
        exist, then lookup in the model.
        """  #
        try:
            attr = getattr(self.get_queryset(), item)
        except AttributeError:
            attr = getattr(self.through, item)

        return attr

    def all(self, clear_cache: bool = False) -> "QuerySet":
        # get_queryset returns already a fresh queryset. Skip making a copy.
        return self.get_queryset()

    def expand_relationship(self, value: Any) -> Any:
        through = self.through

        if isinstance(value, (through, through.proxy_model)):  # type: ignore
            return value
        instance = through.proxy_model(
            **{self.from_foreign_key: self.instance, self.to_foreign_key: value}
        )
        instance.identifying_db_fields = [self.from_foreign_key, self.to_foreign_key]  # type: ignore
        return instance

    def stage(self, *children: "BaseModelType") -> None:
        for child in children:
            if not isinstance(
                child,
                (self.to, self.to.proxy_model, self.through, self.through.proxy_model, dict),  # type: ignore
            ):
                raise RelationshipIncompatible(
                    f"The child is not from the types '{self.to.__name__}', '{self.through.__name__}'."
                )
            self.refs.append(self.expand_relationship(child))

    async def add(self, child: "BaseModelType") -> Optional["BaseModelType"]:
        """
        Adds a child to the model as a list

        . Validates the type of the child being added to the relationship and raises error for
        if the type is wrong.
        . Checks if the middle table already contains the record being added. Raises error if yes.
        """
        if not isinstance(
            child,
            (self.to, self.to.proxy_model, self.through, self.through.proxy_model, dict),  # type: ignore
        ):
            raise RelationshipIncompatible(
                f"The child is not from the types '{self.to.__name__}', '{self.through.__name__}'."
            )
        child = self.expand_relationship(child)
        # try saving intermediate model. If it fails another instance already exists and return None
        try:
            return await child.save(force_insert=True)
        except IntegrityError:
            pass
        return None

    async def remove(self, child: Optional["BaseModelType"] = None) -> None:
        """Removes a child from the list of many to many.

        . Validates if there is a relationship between the entities.
        . Removes the field if there is
        """
        if self.reverse:
            fk = self.through.meta.fields[self.from_foreign_key]
        else:
            fk = self.through.meta.fields[self.to_foreign_key]
        if child is None:
            if fk.unique:
                try:
                    child = await self.get()
                except ObjectNotFound:
                    raise RelationshipNotFound(detail="no child found") from None
            else:
                raise RelationshipNotFound(detail="no child specified")
        if not isinstance(
            child,
            (self.to, self.to.proxy_model, self.through, self.through.proxy_model),  # type: ignore
        ):
            raise RelationshipIncompatible(
                f"The child is not from the types '{self.to.__name__}', '{self.through.__name__}'."
            )
        child = cast("BaseModelType", self.expand_relationship(child))
        count = await child.query.filter(*child.identifying_clauses()).count()
        if count == 0:
            raise RelationshipNotFound(
                detail=f"There is no relationship between '{self.from_foreign_key}' and '{self.to_foreign_key}: {getattr(child,self.to_foreign_key).pk}'."
            )
        else:
            await child.delete()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"{self.through.__name__}"

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> ManyRelationProtocol:
        self.instance = instance
        return self


class SingleRelation(ManyRelationProtocol):
    """
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(
        self,
        *,
        to_foreign_key: str,
        to: type["BaseModelType"],
        embed_parent: Optional[tuple[str, str]] = None,
        refs: Any = (),
        instance: Optional["BaseModelType"] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.to = to
        self.instance = instance
        self.to_foreign_key = to_foreign_key
        self.embed_parent = embed_parent
        self.refs: Sequence[BaseModelType] = []
        if not isinstance(refs, Sequence):
            refs = [refs]
        self.stage(*refs)

    def get_queryset(self) -> "QuerySet":
        # we need to check tenant every request
        queryset = self.to.meta.managers["query_related"].get_queryset()
        fk = self.to.meta.fields[self.to_foreign_key]
        assert self.instance, "instance not initialized"
        query = {}
        for column_name in fk.get_column_names():
            related_name = fk.from_fk_field_name(fk.name, column_name)
            query[related_name] = getattr(self.instance, related_name)
        queryset = queryset.filter(**{self.to_foreign_key: query})
        # now set embed_parent
        queryset.embed_parent = self.embed_parent
        # only apply if relationship field. Add fallback for empty embed_parent
        if self.embed_parent and isinstance(
            fk.owner.meta.fields[self.embed_parent[0].split("__", 1)[0]],
            RelationshipField,
        ):
            queryset.embed_parent_filters = queryset.embed_parent
        return queryset

    def all(self, clear_cache: bool = False) -> "QuerySet":
        # get_queryset returns already a fresh queryset. Skip making a copy.
        return self.get_queryset()

    def expand_relationship(self, value: Any) -> Any:
        target = self.to

        if isinstance(value, (target, target.proxy_model)):  # type: ignore
            return value
        related_columns = self.to.meta.fields[self.to_foreign_key].related_columns.keys()
        if len(related_columns) == 1 and not isinstance(value, (dict, BaseModel)):
            value = {next(iter(related_columns)): value}
        instance = target.proxy_model(**value)
        instance.identifying_db_fields = related_columns  # type: ignore
        return instance

    def stage(self, *children: "BaseModelType") -> None:
        for child in children:
            if not isinstance(child, (self.to, self.to.proxy_model, dict)):  # type: ignore
                raise RelationshipIncompatible(
                    f"The child is not from the types '{self.to.__name__}', '{self.through.__name__}'."
                )
            self.refs.append(self.expand_relationship(child))

    def __getattr__(self, item: Any) -> Any:
        """
        Gets the attribute from the queryset and if it does not
        exist, then lookup in the model.
        """
        try:
            attr = getattr(self.get_queryset(), item)
        except AttributeError:
            attr = getattr(self.to, item)

        return attr

    async def save_related(self) -> None:
        while self.refs:
            await self.add(self.refs.pop())

    async def add(self, child: "BaseModelType") -> Optional["BaseModelType"]:
        """
        Adds a child to the model as a list

        . Validates the type of the child being added to the relationship and raises error for
        if the type is wrong.
        . Checks if the middle table already contains the record being added. Raises error if yes.
        """
        if not isinstance(child, (self.to, self.to.proxy_model, dict)):  # type: ignore
            raise RelationshipIncompatible(f"The child is not from the type '{self.to.__name__}'.")
        child = self.expand_relationship(child)
        await child.save(values={self.to_foreign_key: self.instance})
        return child

    async def remove(self, child: Optional["BaseModelType"] = None) -> None:
        """Removes a child from the list of one to many.

        . Validates if there is a relationship between the entities.
        . Removes the field if there is
        """
        fk = self.to.meta.fields[self.to_foreign_key]
        if child is None:
            if fk.unique:
                try:
                    child = await self.get()
                except ObjectNotFound:
                    raise RelationshipNotFound(detail="no child found") from None
            else:
                raise RelationshipNotFound(detail="no child specified")
        if not isinstance(child, self.to):
            raise RelationshipIncompatible(f"The child is not from the type '{self.to.__name__}'.")

        await child.save(values={self.to_foreign_key: None})

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"{self.to.__name__}"

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> ManyRelationProtocol:
        self.instance = instance
        return self


class VirtualCascadeDeletionSingleRelation(SingleRelation):
    async def post_delete_callback(self, instance: "BaseModelType") -> None:
        await self.delete()



from .db import DatabaseMixin
from .generics import DeclarativeMixin
from .reflection import ReflectedModelMixin
from .row import ModelRowMixin

__all__ = ["DeclarativeMixin", "ModelRowMixin", "ReflectedModelMixin", "DatabaseMixin"]



import os
import sys
import typing

from edgy import Registry
from edgy.cli.operations.shell.utils import import_objects
from edgy.conf import settings
from edgy.core.terminal import Print

printer = Print()


def vi_mode() -> typing.Any:
    editor = os.environ.get("EDITOR")
    if not editor:
        return False
    editor = os.path.basename(editor)
    return editor.startswith("vi") or editor.endswith("vim")


def get_ptpython(app: typing.Any, registry: Registry, options: typing.Any = None) -> typing.Any:
    """Gets the PTPython shell.

    Loads the initial configurations from the main Edgy settings
    and boots up the kernel.
    """
    try:
        from ptpython.repl import embed, run_config

        def run_ptpython() -> None:
            imported_objects = import_objects(app, registry)
            history_filename = os.path.expanduser("~/.ptpython_history")

            config_file = os.path.expanduser(settings.ptpython_config_file)
            if not os.path.exists(config_file):
                embed(
                    globals=imported_objects,
                    history_filename=history_filename,
                    vi_mode=vi_mode(),
                )
            else:
                embed(
                    globals=imported_objects,
                    history_filename=history_filename,
                    vi_mode=vi_mode(),
                    configure=run_config,
                )

    except (ModuleNotFoundError, ImportError):
        error = "You must have ptpython installed to run this. Run `pip install ptpython`"
        printer.write_error(error)
        sys.exit(1)

    return run_ptpython



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Awaitable, Generator, Iterable, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
)

from edgy.types import Undefined

if TYPE_CHECKING:
    import sqlalchemy
    from databasez.core.transaction import Transaction

    from edgy.core.connection import Database
    from edgy.core.db.models import BaseModelType

# Create a var type for the Edgy Model
EdgyModel = TypeVar("EdgyModel", bound="BaseModelType")
EdgyEmbedTarget = TypeVar("EdgyEmbedTarget")

tables_and_models_type = dict[str, tuple["sqlalchemy.Table", type["BaseModelType"]]]


class QuerySetType(ABC, Generic[EdgyEmbedTarget, EdgyModel]):
    __slots__ = ("model_class",)
    model_class: type[EdgyModel]

    def __init__(self, model_class: type[EdgyModel]) -> None:
        self.model_class = model_class

    @abstractmethod
    async def build_where_clause(
        self, _: Any = None, tables_and_models: Optional[tables_and_models_type] = None
    ) -> Any: ...

    @abstractmethod
    def filter(
        self,
        *clauses: Union[
            "sqlalchemy.sql.expression.BinaryExpression",
            Callable[
                ["QuerySetType"],
                Union[
                    "sqlalchemy.sql.expression.BinaryExpression",
                    Awaitable["sqlalchemy.sql.expression.BinaryExpression"],
                ],
            ],
            dict[str, Any],
            "QuerySetType",
        ],
        **kwargs: Any,
    ) -> "QuerySetType": ...

    @abstractmethod
    def all(self, clear_cache: bool = False) -> "QuerySetType": ...

    @abstractmethod
    def or_(
        self,
        *clauses: Union[
            "sqlalchemy.sql.expression.BinaryExpression",
            Callable[
                ["QuerySetType"],
                Union[
                    "sqlalchemy.sql.expression.BinaryExpression",
                    Awaitable["sqlalchemy.sql.expression.BinaryExpression"],
                ],
            ],
            "QuerySetType",
        ],
        **kwargs: Any,
    ) -> "QuerySetType":
        """
        Filters the QuerySet by the OR operand.
        """

    @abstractmethod
    def local_or(
        self,
        *clauses: Union[
            "sqlalchemy.sql.expression.BinaryExpression",
            Callable[
                ["QuerySetType"],
                Union[
                    "sqlalchemy.sql.expression.BinaryExpression",
                    Awaitable["sqlalchemy.sql.expression.BinaryExpression"],
                ],
            ],
            "QuerySetType",
        ],
        **kwargs: Any,
    ) -> "QuerySetType":
        """
        Filters the QuerySet by the OR operand.
        """

    @abstractmethod
    def and_(
        self,
        *clauses: Union[
            "sqlalchemy.sql.expression.BinaryExpression",
            Callable[
                ["QuerySetType"],
                Union[
                    "sqlalchemy.sql.expression.BinaryExpression",
                    Awaitable["sqlalchemy.sql.expression.BinaryExpression"],
                ],
            ],
            dict[str, Any],
            "QuerySetType",
        ],
        **kwargs: Any,
    ) -> "QuerySetType":
        """
        Filters the QuerySet by the AND operand. Alias of filter.
        """

    @abstractmethod
    def not_(
        self,
        *clauses: Union[
            "sqlalchemy.sql.expression.BinaryExpression",
            Callable[
                ["QuerySetType"],
                Union[
                    "sqlalchemy.sql.expression.BinaryExpression",
                    Awaitable["sqlalchemy.sql.expression.BinaryExpression"],
                ],
            ],
            dict[str, Any],
            "QuerySetType",
        ],
        **kwargs: Any,
    ) -> "QuerySetType":
        """
        Filters the QuerySet by the NOT operand. Alias of exclude.
        """
        raise NotImplementedError()

    @abstractmethod
    def exclude(
        self,
        *clauses: Union[
            "sqlalchemy.sql.expression.BinaryExpression",
            Callable[
                ["QuerySetType"],
                Union[
                    "sqlalchemy.sql.expression.BinaryExpression",
                    Awaitable["sqlalchemy.sql.expression.BinaryExpression"],
                ],
            ],
            dict[str, Any],
            "QuerySetType",
        ],
        **kwargs: Any,
    ) -> "QuerySetType": ...

    @abstractmethod
    def lookup(self, term: Any) -> "QuerySetType": ...

    @abstractmethod
    def order_by(self, *columns: str) -> "QuerySetType": ...

    @abstractmethod
    def reverse(self) -> "QuerySetType": ...

    @abstractmethod
    def limit(self, limit_count: int) -> "QuerySetType": ...

    @abstractmethod
    def offset(self, offset: int) -> "QuerySetType": ...

    @abstractmethod
    def group_by(self, *group_by: str) -> "QuerySetType": ...

    @abstractmethod
    def distinct(self, *distinct_on: Sequence[str]) -> "QuerySetType": ...

    @abstractmethod
    def select_related(self, *related: str) -> "QuerySetType": ...

    @abstractmethod
    def only(self, *fields: str) -> "QuerySetType": ...

    @abstractmethod
    def defer(self, *fields: str) -> "QuerySetType": ...

    @abstractmethod
    async def exists(self) -> bool: ...

    @abstractmethod
    async def count(self) -> int: ...

    @abstractmethod
    async def get_or_none(self, **kwargs: Any) -> Union[EdgyEmbedTarget, None]: ...

    @abstractmethod
    async def get(self, **kwargs: Any) -> EdgyEmbedTarget: ...

    @abstractmethod
    async def first(self) -> Union[EdgyEmbedTarget, None]: ...

    @abstractmethod
    async def last(self) -> Union[EdgyEmbedTarget, None]: ...

    @abstractmethod
    async def create(self, *args: Any, **kwargs: Any) -> EdgyEmbedTarget: ...

    @abstractmethod
    async def bulk_create(self, objs: Iterable[Union[dict[str, Any], EdgyModel]]) -> None: ...

    @abstractmethod
    async def bulk_update(self, objs: Sequence[EdgyModel], fields: list[str]) -> None: ...

    @abstractmethod
    async def delete(self) -> None: ...

    @abstractmethod
    async def update(self, **kwargs: Any) -> None: ...

    @abstractmethod
    async def values(
        self,
        fields: Union[Sequence[str], str, None],
        exclude: Union[Sequence[str], set[str]],
        exclude_none: bool,
        flatten: bool,
        **kwargs: Any,
    ) -> list[Any]: ...

    @abstractmethod
    async def values_list(
        self,
        fields: Union[Sequence[str], str, None],
        exclude: Union[Sequence[str], set[str]],
        exclude_none: bool,
        flat: bool,
    ) -> list[Any]: ...

    @abstractmethod
    async def get_or_create(
        self,
        defaults: Union[dict[str, Any], Any, None] = None,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[EdgyEmbedTarget, bool]: ...

    @abstractmethod
    async def update_or_create(
        self, defaults: Union[dict[str, Any], Any, None] = None, *args: Any, **kwargs: Any
    ) -> tuple[EdgyEmbedTarget, bool]: ...

    @abstractmethod
    async def contains(self, instance: "BaseModelType") -> bool: ...

    @abstractmethod
    def transaction(self, *, force_rollback: bool = False, **kwargs: Any) -> "Transaction":
        """Return database transaction for the assigned database."""

    @abstractmethod
    def using(
        self,
        *,
        database: Union[str, Any, None, "Database"] = Undefined,
        schema: Union[str, Any, None, Literal[False]] = Undefined,
    ) -> "QuerySetType": ...

    @abstractmethod
    def __await__(self) -> Generator[Any, None, list[EdgyEmbedTarget]]: ...

    @abstractmethod
    async def __aiter__(self) -> AsyncIterator[EdgyEmbedTarget]: ...


QueryType = QuerySetType



import asyncio
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Optional, Union

import sqlalchemy
from sqlalchemy.exc import DBAPIError, ProgrammingError

from edgy.core.connection.database import Database
from edgy.exceptions import SchemaError

if TYPE_CHECKING:
    from edgy import Registry


class Schema:
    """
    All the schema operations object.

    All the operations regarding a schema are placed in one object
    """

    _default_schema: Optional[str]

    def __init__(self, registry: "Registry") -> None:
        self.registry = registry

    @property
    def database(self) -> "Database":
        return self.registry.database

    def get_default_schema(self) -> Optional[str]:
        """
        Returns the default schema which is usually None
        """
        if not hasattr(self, "_default_schema"):
            self._default_schema = self.database.url.sqla_url.get_dialect(True).default_schema_name
        return self._default_schema

    async def activate_schema_path(
        self, database: Database, schema: str, is_shared: bool = True
    ) -> None:
        # INSECURE, but not used by default. Add warning
        # TODO: remove when there are no users of the method
        warnings.warn(
            "`activate_schema_path` is dangerous because the schema is not properly escaped and deprecated.",
            DeprecationWarning,
            stacklevel=2,
        )
        path = (
            f"SET search_path TO {schema}, shared;"
            if is_shared
            else f"SET search_path TO {schema};"
        )
        expression = sqlalchemy.text(path)
        await database.execute(expression)

    async def create_schema(
        self,
        schema: str,
        if_not_exists: bool = False,
        init_models: bool = False,
        init_tenant_models: bool = False,
        update_cache: bool = True,
        databases: Sequence[Union[str, None]] = (None,),
    ) -> None:
        """
        Creates a model schema if it does not exist.
        """
        tenant_tables: list[sqlalchemy.Table] = []
        if init_models:
            for model_class in self.registry.models.values():
                model_class.table_schema(schema=schema, update_cache=update_cache)
        if init_tenant_models and init_models:
            for model_class in self.registry.tenant_models.values():
                model_class.table_schema(schema=schema, update_cache=update_cache)
        elif init_tenant_models:
            for model_class in self.registry.tenant_models.values():
                tenant_tables.append(
                    model_class.table_schema(schema=schema, update_cache=update_cache)
                )

        def execute_create(connection: sqlalchemy.Connection, name: Optional[str]) -> None:
            try:
                connection.execute(
                    sqlalchemy.schema.CreateSchema(name=schema, if_not_exists=if_not_exists)
                )
            except ProgrammingError as e:
                raise SchemaError(detail=e.orig.args[0]) from e
            for table in tenant_tables:
                table.create(connection, checkfirst=if_not_exists)
            if init_models:
                self.registry.metadata_by_name[name].create_all(
                    connection, checkfirst=if_not_exists
                )

        ops = []
        for database_name in databases:
            db = (
                self.registry.database
                if database_name is None
                else self.registry.extra[database_name]
            )
            # don't warn here about inperformance
            async with db as db:
                with db.force_rollback(False):
                    ops.append(db.run_sync(execute_create, database_name))
        await asyncio.gather(*ops)

    async def drop_schema(
        self,
        schema: str,
        cascade: bool = False,
        if_exists: bool = False,
        databases: Sequence[Union[str, None]] = (None,),
    ) -> None:
        """
        Drops an existing model schema.
        """

        def execute_drop(connection: sqlalchemy.Connection) -> None:
            try:
                connection.execute(
                    sqlalchemy.schema.DropSchema(name=schema, cascade=cascade, if_exists=if_exists)
                )
            except DBAPIError as e:
                raise SchemaError(detail=e.orig.args[0]) from e

        ops = []

        for database_name in databases:
            db = (
                self.registry.database
                if database_name is None
                else self.registry.extra[database_name]
            )
            # don't warn here about inperformance
            async with db as db:
                with db.force_rollback(False):
                    ops.append(db.run_sync(execute_drop))
        await asyncio.gather(*ops)

    async def get_metadata_of_all_schemes(
        self, database: Database, *, no_reflect: bool = False
    ) -> tuple[sqlalchemy.MetaData, list[str]]:
        tablenames = self.registry.get_tablenames()

        async with database as database:
            list_schemes: list[str] = []
            metadata = sqlalchemy.MetaData()

            def wrapper(connection: sqlalchemy.Connection) -> None:
                nonlocal list_schemes
                inspector = sqlalchemy.inspect(connection)
                default_schema_name = inspector.default_schema_name
                list_schemes = [
                    "" if default_schema_name == schema else schema
                    for schema in inspector.get_schema_names()
                ]
                if not no_reflect:
                    for schema in list_schemes:
                        metadata.reflect(
                            connection, schema=schema, only=lambda name, _: name in tablenames
                        )

            await database.run_sync(wrapper)
            return metadata, list_schemes

    async def get_schemes_tree(
        self, *, no_reflect: bool = False
    ) -> dict[Union[str, None], tuple[str, sqlalchemy.MetaData, list[str]]]:
        schemes_tree: dict[Union[str, None], tuple[str, sqlalchemy.MetaData, list[str]]] = {
            None: (
                str(self.database.url),
                *(
                    await self.get_metadata_of_all_schemes(
                        self.registry.database, no_reflect=no_reflect
                    )
                ),
            )
        }
        for key, val in self.registry.extra.items():
            schemes_tree[key] = (
                str(val.url),
                *(await self.get_metadata_of_all_schemes(val, no_reflect=no_reflect)),
            )
        return schemes_tree



from typing import TYPE_CHECKING, Any

from edgy.core.db.context_vars import CURRENT_PHASE
from edgy.core.db.fields.factories import FieldFactory

if TYPE_CHECKING:
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.types import BaseModelType


class ExcludeField(FieldFactory, type[None]):
    """
    Meta field that masks fields
    """

    field_type: Any = Any

    def __new__(  # type: ignore
        cls,
        **kwargs: Any,
    ) -> "BaseFieldType":
        kwargs["exclude"] = True
        kwargs["null"] = True
        kwargs["primary_key"] = False
        return super().__new__(cls, **kwargs)

    @classmethod
    def clean(
        cls,
        obj: "BaseFieldType",
        name: str,
        value: Any,
        for_query: bool = False,
        original_fn: Any = None,
    ) -> dict[str, Any]:
        """remove any value from input."""
        return {}

    @classmethod
    def to_model(
        cls,
        obj: "BaseFieldType",
        name: str,
        value: Any,
        original_fn: Any = None,
    ) -> dict[str, Any]:
        """remove any value from input and raise when setting an attribute."""
        phase = CURRENT_PHASE.get()
        if phase == "set":
            raise ValueError("field is excluded")
        return {}

    @classmethod
    def __get__(
        cls,
        obj: "BaseFieldType",
        instance: "BaseModelType",
        owner: Any = None,
        original_fn: Any = None,
    ) -> None:
        raise ValueError("field is excluded")



import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional, cast

from edgy.core.db.fields.base import RelationshipField
from edgy.core.db.models.utils import apply_instance_extras
from edgy.core.db.querysets.prefetch import Prefetch, check_prefetch_collision
from edgy.core.db.relationships.utils import crawl_relationship
from edgy.exceptions import QuerySetError

if TYPE_CHECKING:  # pragma: no cover
    from sqlalchemy import Table
    from sqlalchemy.engine.result import Row

    from edgy import Database, Model
    from edgy.core.db.models.types import BaseModelType


class ModelRowMixin:
    """
    Builds a row for a specific model
    """

    @classmethod
    def can_load_from_row(cls: type["Model"], row: "Row", table: "Table") -> bool:
        """Check if a model_class can be loaded from a row for the table."""

        return bool(
            cls.meta.registry
            and not cls.meta.abstract
            and all(row._mapping.get(f"{table.name}_{col}") is not None for col in cls.pkcolumns)
        )

    @classmethod
    async def from_sqla_row(
        cls: type["Model"],
        row: "Row",
        # contain the mappings used for select
        tables_and_models: dict[str, tuple["Table", type["BaseModelType"]]],
        select_related: Optional[Sequence[Any]] = None,
        prefetch_related: Optional[Sequence["Prefetch"]] = None,
        only_fields: Sequence[str] = None,
        is_defer_fields: bool = False,
        exclude_secrets: bool = False,
        using_schema: Optional[str] = None,
        database: Optional["Database"] = None,
        prefix: str = "",
        old_select_related_value: Optional["Model"] = None,
    ) -> Optional["Model"]:
        """
        Class method to convert a SQLAlchemy Row result into a EdgyModel row type.

        Looping through select_related fields if the query comes from a select_related operation.
        Validates if exists the select_related and related_field inside the models.

        When select_related and related_field exist for the same field being validated, the related
        field is ignored as it won't override the value already collected from the select_related.

        If there is no select_related, then goes through the related field where it **should**
        only return the instance of the the ForeignKey with the ID, making it lazy loaded.

        :return: Model class.
        """
        item: dict[str, Any] = {}
        select_related = select_related or []
        prefetch_related = prefetch_related or []
        secret_columns: set[str] = set()
        if exclude_secrets:
            for name in cls.meta.secret_fields:
                secret_columns.update(cls.meta.field_to_column_names[name])

        for related in select_related:
            field_name = related.split("__", 1)[0]
            try:
                field = cls.meta.fields[field_name]
            except KeyError:
                raise QuerySetError(
                    detail=f'Selected field "{field_name}cast("Model", " does not exist on {cls}.'
                ) from None
            if isinstance(field, RelationshipField):
                model_class, _, remainder = field.traverse_field(related)
            else:
                raise QuerySetError(
                    detail=f'Selected field "{field_name}" is not a RelationshipField on {cls}.'
                ) from None

            _prefix = field_name if not prefix else f"{prefix}__{field_name}"
            # stop selecting when None. Related models are not available.
            if not model_class.can_load_from_row(
                row,
                tables_and_models[_prefix][0],
            ):
                continue

            if remainder:
                # don't pass table, it is only for the main model_class
                item[field_name] = await model_class.from_sqla_row(
                    row,
                    tables_and_models=tables_and_models,
                    select_related=[remainder],
                    prefetch_related=prefetch_related,
                    exclude_secrets=exclude_secrets,
                    is_defer_fields=is_defer_fields,
                    using_schema=using_schema,
                    database=database,
                    prefix=_prefix,
                    old_select_related_value=item.get(field_name),
                )
            else:
                # don't pass table, it is only for the main model_class
                item[field_name] = await model_class.from_sqla_row(
                    row,
                    tables_and_models=tables_and_models,
                    exclude_secrets=exclude_secrets,
                    is_defer_fields=is_defer_fields,
                    using_schema=using_schema,
                    database=database,
                    prefix=_prefix,
                    old_select_related_value=item.get(field_name),
                )
        # don't overwrite, update with new values and return
        if old_select_related_value:
            for k, v in item.items():
                setattr(old_select_related_value, k, v)
            return old_select_related_value
        table_columns = tables_and_models[prefix][0].columns
        # Populate the related names
        # Making sure if the model being queried is not inside a select related
        # This way it is not overritten by any value
        for related in cls.meta.foreign_key_fields:
            foreign_key = cls.meta.fields[related]
            ignore_related: bool = cls.__should_ignore_related_name(related, select_related)
            if ignore_related or related in cls.meta.secret_fields:
                continue
            if related in item:
                continue

            if exclude_secrets and foreign_key.secret:
                continue
            columns_to_check = foreign_key.get_column_names(related)

            model_related = foreign_key.target

            child_item = {}
            for column_name in columns_to_check:
                column = getattr(table_columns, column_name, None)
                if column_name is None:
                    continue
                columnkeyhash = column_name
                if prefix:
                    columnkeyhash = f"{tables_and_models[prefix][0].name}_{column.key}"

                if columnkeyhash in row._mapping:
                    child_item[foreign_key.from_fk_field_name(related, column_name)] = (
                        row._mapping[columnkeyhash]
                    )
            # Make sure we generate a temporary reduced model
            # For the related fields. We simply chnage the structure of the model
            # and rebuild it with the new fields.
            proxy_model = model_related.proxy_model(**child_item)
            proxy_database = database if model_related.database is cls.database else None
            # don't pass a table. It is not in the row (select related path) and has not an explicit table
            proxy_model = apply_instance_extras(
                proxy_model,
                model_related,
                using_schema,
                database=proxy_database,
            )
            proxy_model.identifying_db_fields = foreign_key.related_columns

            item[related] = proxy_model

        # Check for the only_fields
        # Pull out the regular column values.
        for column in table_columns:
            if (
                only_fields
                and prefix not in only_fields
                and (f"{prefix}__{column.key}" if prefix else column.key) not in only_fields
            ):
                continue
            if column.key in secret_columns:
                continue
            if column.key not in cls.meta.columns_to_field:
                continue
            # set if not of an foreign key with one column
            if column.key in item:
                continue
            columnkeyhash = column.key
            if prefix:
                columnkeyhash = f"{tables_and_models[prefix][0].name}_{columnkeyhash}"

            if columnkeyhash in row._mapping:
                item[column.key] = row._mapping[columnkeyhash]
        model: Model = (
            cls.proxy_model(**item, __phase__="init_db")  # type: ignore
            if exclude_secrets or is_defer_fields or only_fields
            else cls(**item, __phase__="init_db")
        )
        # Apply the schema to the model
        model = apply_instance_extras(
            model,
            cls,
            using_schema,
            database=database,
            table=tables_and_models[prefix][0],
        )

        if prefetch_related:
            # Handle prefetch related fields.
            await cls.__handle_prefetch_related(
                row=row,
                prefix=prefix,
                model=model,
                tables_and_models=tables_and_models,
                prefetch_related=prefetch_related,
            )
        assert model.pk is not None, model
        return model

    @classmethod
    def __should_ignore_related_name(
        cls, related_name: str, select_related: Sequence[str]
    ) -> bool:
        """
        Validates if it should populate the related field if select related is not considered.
        """
        for related_field in select_related:
            fields = related_field.split("__")
            if related_name in fields:
                return True
        return False

    @classmethod
    def create_model_key_from_sqla_row(cls, row: "Row", row_prefix: str = "") -> tuple:
        """
        Build a cache key for the model.
        """
        pk_key_list: list[Any] = [cls.__name__]
        for attr in cls.pkcolumns:
            pk_key_list.append(str(row._mapping[f"{row_prefix}{attr}"]))
        return tuple(pk_key_list)

    @classmethod
    async def __set_prefetch(
        cls,
        row: "Row",
        model: "Model",
        row_prefix: str,
        related: "Prefetch",
    ) -> None:
        model_key = ()
        if related._is_finished:
            # when force_rollback
            # we can only bake after all rows are retrieved
            # this is why it is here
            await related.init_bake(type(model))
            model_key = model.create_model_key()
        if model_key in related._baked_results:
            setattr(model, related.to_attr, related._baked_results[model_key])
        else:
            crawl_result = crawl_relationship(
                model.__class__, related.related_name, traverse_last=True
            )
            if crawl_result.reverse_path is False:
                QuerySetError(
                    detail=("Creating a reverse path is not possible, unidirectional fields used.")
                )
            if crawl_result.cross_db_remainder:
                raise NotImplementedError(
                    "Cannot prefetch from other db yet. Maybe in future this feature will be added."
                )
            queryset = related.queryset
            if related._is_finished:
                assert queryset is not None, "Queryset is not set but _is_finished flag"
            else:
                check_prefetch_collision(model, related)
                if queryset is None:
                    queryset = crawl_result.model_class.query.all()

                queryset = queryset.select_related(cast(str, crawl_result.reverse_path))
            clause = {
                f"{crawl_result.reverse_path}__{pkcol}": row._mapping[f"{row_prefix}{pkcol}"]
                for pkcol in cls.pkcolumns
            }
            setattr(model, related.to_attr, await queryset.filter(clause))

    @classmethod
    async def __handle_prefetch_related(
        cls,
        row: "Row",
        model: "Model",
        prefix: str,
        tables_and_models: dict[str, tuple["Table", type["BaseModelType"]]],
        prefetch_related: Sequence["Prefetch"],
    ) -> None:
        """
        Handles any prefetch related scenario from the model.
        Loads in advance all the models needed for a specific record

        Recursively checks for the related field and validates if there is any conflicting
        attribute. If there is, a `QuerySetError` is raised.
        """

        queries = []

        for related in prefetch_related:
            # Check for conflicting names
            # Check as early as possible
            check_prefetch_collision(model=model, related=related)
            row_prefix = f"{tables_and_models[prefix].name}_" if prefix else ""
            queries.append(
                cls.__set_prefetch(row=row, row_prefix=row_prefix, model=model, related=related)
            )
        if queries:
            await asyncio.gather(*queries)



from typing import Any

from blinker import Signal

from edgy.exceptions import SignalError

post_delete = Signal()
post_save = Signal()
post_update = Signal()
pre_delete = Signal()
pre_save = Signal()
pre_update = Signal()


class Broadcaster(dict):
    def __getattr__(self, item: str) -> Signal:
        return self.setdefault(item, Signal())  # type: ignore

    def __setattr__(self, __name: str, __value: Signal) -> None:
        if not isinstance(__value, Signal):
            raise SignalError(f"{__value} is not valid signal")
        self[__name] = __value

    def set_lifecycle_signals_from(self, namespace: Any, overwrite: bool = True) -> None:
        for name in (
            "post_delete",
            "post_save",
            "post_update",
            "pre_delete",
            "pre_save",
            "pre_update",
        ):
            if overwrite:
                setattr(self, name, getattr(namespace, name))
            else:
                self.setdefault(name, getattr(namespace, name))


__all__ = [
    "Signal",
    "Broadcaster",
    "post_delete",
    "post_save",
    "post_update",
    "pre_delete",
    "pre_save",
    "pre_update",
]



from typing import ClassVar, Optional

import edgy

from .managers import PermissionManager


class BasePermission(edgy.Model):
    users_field_group: ClassVar[str] = "users"
    name: str = edgy.fields.CharField(max_length=100, null=False)
    # model_name would collidate with pydantics private namespace
    # name_model: str = edgy.fields.CharField(max_length=100, null=True)
    # obj = edgy.fields.ForeignKey("ContentType", null=True)
    description: Optional[str] = edgy.fields.ComputedField(  # type: ignore
        getter="get_description",
        setter="set_description",
        # default to name
        fallback_getter=lambda field, instance, owner: instance.name,
    )

    # Important: embed_through must be set for enabling full proxying
    # users = edgy.fields.ManyToMany("User", embed_through=False)
    # groups = edgy.fields.ManyToMany("Group", embed_through=False)

    query = PermissionManager()

    class Meta:
        abstract = True
        # unique_together=[(name, name_model, obj)]



from __future__ import annotations

import copy
import inspect
from typing import ClassVar, cast

from edgy.core.db.models.base import EdgyBaseModel
from edgy.core.db.models.managers import Manager, RedirectManager
from edgy.core.db.models.metaclasses import BaseModelMeta, MetaInfo
from edgy.core.db.models.mixins import (
    DatabaseMixin,
    DeclarativeMixin,
    ModelRowMixin,
    ReflectedModelMixin,
)
from edgy.core.db.models.model_proxy import ProxyModel
from edgy.core.utils.models import generify_model_fields


class Model(
    ModelRowMixin, DeclarativeMixin, DatabaseMixin, EdgyBaseModel, metaclass=BaseModelMeta
):
    """
    Representation of an Edgy `Model`.

    This also means it can generate declarative SQLAlchemy models
    from anywhere by calling the `Model.declarative()` function.

    **Example**

    ```python
    import edgyBaseFieldType
    from edgy import Database, Registry

    database = Database("sqlite:///db.sqlite")
    models = Registry(database=database)


    class User(edgy.Model):
        '''
        The User model to be created in the database as a table
        If no name is provided the in Meta class, it will generate
        a "users" table for you.
        '''

        id: int = edgy.IntegerField(primary_key=True, autoincrement=True)
        is_active: bool = edgy.BooleanField(default=False)

        class Meta:
            registry = models
    ```
    """

    query: ClassVar[Manager] = Manager()
    query_related: ClassVar[RedirectManager] = RedirectManager(redirect_name="query")
    # registry = False, stops the retrieval of the registry from base classes
    meta: ClassVar[MetaInfo] = MetaInfo(None, abstract=True, registry=False)

    class Meta:
        abstract = True

    @classmethod
    def generate_proxy_model(cls) -> type[Model]:
        """
        Generates a proxy model for each model. This proxy model is a simple
        shallow copy of the original model being generated.
        """
        fields = {key: copy.copy(field) for key, field in cls.meta.fields.items()}

        class MethodHolder(Model):
            pass

        ignore = set(dir(MethodHolder))

        for key in dir(cls):
            if key in ignore or key.startswith("_"):
                continue
            val = inspect.getattr_static(cls, key)
            if inspect.isfunction(val):
                setattr(MethodHolder, key, val)

        proxy_model = ProxyModel(
            name=cls.__name__,
            module=cls.__module__,
            metadata=cls.meta,
            definitions=fields,
            bases=(MethodHolder,),
        )

        proxy_model.build()
        generify_model_fields(cast(type[EdgyBaseModel], proxy_model.model))
        return cast(type[Model], proxy_model.model)


class ReflectModel(ReflectedModelMixin, Model):
    """
    Reflect on async engines is not yet supported, therefore, we need to make a sync_engine
    call.
    """

    class Meta:
        abstract = True
        # registry = False, stops the retrieval of the registry from base classes
        registry = False



import functools
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional, cast

from edgy.core.db.context_vars import CURRENT_INSTANCE
from edgy.core.db.fields.base import RelationshipField
from edgy.core.db.fields.foreign_keys import BaseForeignKeyField
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:
    from edgy.core.connection.database import Database
    from edgy.core.db.models.types import BaseModelType


class RelatedField(RelationshipField):
    """
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """

    def __init__(
        self,
        *,
        foreign_key_name: str,
        related_from: type["BaseModelType"],
        **kwargs: Any,
    ) -> None:
        self.foreign_key_name = foreign_key_name
        self.related_from = related_from
        super().__init__(
            inherit=False,
            exclude=True,
            field_type=Any,
            annotation=Any,
            column_type=None,
            null=True,
            **kwargs,
        )
        if self.foreign_key.relation_has_post_delete_callback:
            self.post_delete_callback = self._notset_post_delete_callback

    @property
    def related_to(self) -> type["BaseModelType"]:
        return self.owner

    @property
    def related_name(self) -> str:
        return self.name

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        """
        Meta field
        """
        instance = cast("BaseModelType", CURRENT_INSTANCE.get())
        if isinstance(value, ManyRelationProtocol):
            return {field_name: value}
        if instance:
            relation_instance = self.__get__(instance)
            if not isinstance(value, Sequence):
                value = [value]
            relation_instance.stage(*value)
        else:
            relation_instance = self.get_relation(refs=value)
        return {field_name: relation_instance}

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> ManyRelationProtocol:
        if not instance:
            raise ValueError("missing instance")

        if instance.__dict__.get(self.name, None) is None:
            instance.__dict__[self.name] = self.get_relation()
        if instance.__dict__[self.name].instance is None:
            instance.__dict__[self.name].instance = instance
        return instance.__dict__[self.name]  # type: ignore

    @functools.cached_property
    def foreign_key(self) -> BaseForeignKeyField:
        return cast(BaseForeignKeyField, self.related_from.meta.fields[self.foreign_key_name])

    def traverse_field(self, path: str) -> tuple[Any, str, str]:
        return self.foreign_key.reverse_traverse_field(path)

    def get_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        return self.foreign_key.get_relation(**kwargs)

    def is_cross_db(self, owner_database: Optional["Database"] = None) -> bool:
        if owner_database is None:
            owner_database = self.owner.database
        return str(owner_database.url) != str(self.foreign_key.owner.database.url)

    @property
    def is_m2m(self) -> bool:
        return self.foreign_key.is_m2m

    def clean(self, name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        return self.foreign_key.reverse_clean(name, value, for_query=for_query)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"({self.related_to.__name__}={self.related_name})"

    async def post_save_callback(
        self, value: ManyRelationProtocol, instance: "BaseModelType", force_insert: bool
    ) -> None:
        await value.save_related()

    async def _notset_post_delete_callback(
        self, value: ManyRelationProtocol, instance: "BaseModelType"
    ) -> None:
        if hasattr(value, "post_delete_callback"):
            await value.post_delete_callback(instance)



from typing import TYPE_CHECKING, Any, ClassVar

from pydantic import BaseModel, ConfigDict

from edgy.core.events import is_async_callable
from edgy.core.marshalls.config import ConfigMarshall
from edgy.core.marshalls.fields import BaseMarshallField
from edgy.core.marshalls.helpers import MarshallFieldMapping
from edgy.core.marshalls.metaclasses import MarshallMeta
from edgy.core.utils.sync import run_sync

if TYPE_CHECKING:
    from edgy.core.db.models.model import Model


class BaseMarshall(BaseModel, metaclass=MarshallMeta):
    """
    Base for all the marshalls of Edgy.
    """

    marshall_config: ClassVar[ConfigMarshall]
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    __show_pk__: ClassVar[bool] = False
    __custom_fields__: ClassVar[dict[str, BaseMarshallField]] = {}

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        self._instance: Model = self._setup()

    def _setup(self) -> "Model":
        """
        Assemble the Marshall object with all the given details.

        Returns:
            Model or None: The assembled model instance, or None if the assembly fails.
        """
        data = self.model_dump(exclude={"id"})
        data["__show_pk__"] = self.__show_pk__
        instance: Model = self.marshall_config["model"](**data)  # type: ignore
        self._resolve_serializer(instance=instance)
        return instance

    @property
    def instance(self) -> "Model":
        return self._instance

    @instance.setter
    def instance(self, value: "Model") -> None:
        self._instance = value

    def _resolve_serializer(self, instance: "Model") -> "BaseMarshall":
        """
        Resolve serializer fields and populate them with data from the provided instance.

        Args:
            instance (Model or None, optional): The instance to extract data from. Defaults to None.

        Returns:
            BaseMarshall: The resolved serializer instance.
        """
        # Dump model data excluding fields extracted above
        data = self.model_dump(exclude=set(self.__custom_fields__.keys()))

        # Iterate over fields to populate them with data
        for name, field in self.__custom_fields__.items():
            if field.source and field.source in data:
                setattr(self, name, getattr(instance, field.source))
            elif field.source and not field.__is_method__:
                # For primary key exceptions
                if name in instance.pknames:
                    attribute = getattr(instance, name)
                else:
                    attribute = getattr(instance, field.source)

                # If attribute is callable, execute it and set the value
                if callable(attribute):
                    value = run_sync(attribute()) if is_async_callable(attribute) else attribute()
                else:
                    value = attribute
                setattr(self, name, value)
            elif field.__is_method__:
                value = self._get_method_value(name, instance)
                setattr(self, name, value)
        return self

    def _get_method_value(self, name: str, instance: "Model") -> Any:
        """
        Retrieve the value for a method-based field.

        Args:
            name (str): The name of the method-based field.
            instance (Model): The instance to retrieve the value from.

        Returns:
            Optional: The value retrieved from the method-based field.
        """
        func_name: str = f"get_{name}"
        func = getattr(self, func_name)
        return func(instance)

    def _handle_primary_key(self, instance: "Model") -> None:
        """
        Handles the field of a the primary key if present in the
        fields.
        """
        # Dump model data excluding fields extracted above
        data = self.model_dump(exclude=set(self.__custom_fields__.keys()))

        pk_attribute_in_data = False
        for pk_attribute in instance.pknames:
            if pk_attribute in data:
                pk_attribute_in_data = True
                break

        if pk_attribute_in_data:
            for pk_attribute in instance.pknames:
                # bypass __setattr__ method
                object.__setattr__(self, pk_attribute, getattr(instance, pk_attribute))

    @property
    def fields(self) -> MarshallFieldMapping:
        """
        Returns all the fields of the Marshall.
        """
        fields = MarshallFieldMapping(self)
        # copy fields from model_fields
        for k, v in self.model_fields.items():
            fields[k] = v
        return fields

    async def save(self) -> Any:
        """
        This save is not actually performing any specific action,
        in fact, its using the Edgy saving method and nothing else.

        We add the save here to make sure we have compatibility with
        the interface of Edgy.

        This is optional! This should only be called if you indeed
        want to persist the object in the database and have access to a
        primary key.

        !!! Tip
            All the field and model validations **are still performed**
            in the model level, not in a marshall level.
        """
        instance = await self.instance.save()
        self._handle_primary_key(instance=instance)
        return self


class Marshall(BaseMarshall):
    """
    Model marshall where the `__model__` is required.
    """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"



"""
All functional common to Edgy
"""

from typing import Any, Union

from pydantic.fields import FieldInfo

from edgy.core.db.fields.types import BaseFieldType

# deprecate, it can lead to circular imports and is an one-liner anyway
# bypass __setattr__ method
edgy_setattr = object.__setattr__


def extract_field_annotations_and_defaults(
    attrs: dict[Any, Any], base_type: Union[type[FieldInfo], type[BaseFieldType]] = BaseFieldType
) -> tuple[dict[Any, Any], dict[Any, Any]]:
    """
    Extracts annotations from class namespace dict and triggers
    extraction of ormar model_fields.
    """
    key = "__annotations__"
    attrs[key] = attrs.get(key, {})
    attrs, model_fields = populate_pydantic_default_values(attrs, base_type)
    return attrs, model_fields


def get_model_fields(
    attrs: Union[dict, Any], base_type: Union[type[FieldInfo], type[BaseFieldType]] = BaseFieldType
) -> dict:
    """
    Gets all the fields in current model class that are Edgy Fields.
    """
    return {k: v for k, v in attrs.items() if isinstance(v, base_type)}


def populate_pydantic_default_values(
    attrs: dict, base_type: Union[type[FieldInfo], type[BaseFieldType]] = BaseFieldType
) -> tuple[dict, dict]:
    """
    Making sure the fields from Edgy are the ones being validated by Edgy models
    and delegates the validations from pydantic to that functionality.
    """
    model_fields = {}
    potential_fields = {}

    potential_fields.update(get_model_fields(attrs, base_type))
    for field_name, field in potential_fields.items():
        field.name = field_name
        original_type = getattr(field, "__original_type__", None)

        default_type = field.field_type if not field.null else Union[field.field_type, None]
        overwrite_type = original_type if field.field_type != original_type else None
        field.annotation = overwrite_type or default_type
        model_fields[field_name] = field
        attrs["__annotations__"][field_name] = overwrite_type or default_type
    return attrs, model_fields
